from youseedee import ucd_data, database
from .SyllabicShaper import SyllabicShaper
from .IndicShaperData import make_syllable_machine
from fontFeatures.shaperLib.Buffer import BufferItem
import unicodedata
from collections import OrderedDict

states = OrderedDict(
    h="H | HVM | Sk",
    consonant_modifiers="CMAbv* CMBlw* ((h B | SUB) CMAbv? CMBlw*)*",
    medial_consonants="MPre? MAbv? MBlw? MPst?",
    dependent_vowels="VPre* VAbv* VBlw* VPst*",
    vowel_modifiers="HVM? VMPre* VMAbv* VMBlw* VMPst*",
    final_consonants="FAbv* FBlw* FPst*",
    final_modifiers="FMAbv* FMBlw* | FMPst?",
    complex_syllable_start="(R | CS)? (B | GB)",
    complex_syllable_middle="consonant_modifiers medial_consonants dependent_vowels vowel_modifiers (Sk B)*",
    complex_syllable_tail="complex_syllable_middle final_consonants final_modifiers",
    number_joiner_terminated_cluster_tail="(HN N)* HN",
    numeral_cluster_tail="(HN N)+",
    symbol_cluster_tail="SMAbv+ SMBlw* | SMBlw+",
    virama_terminated_cluster="complex_syllable_start consonant_modifiers h",
    sakot_terminated_cluster="complex_syllable_start complex_syllable_middle Sk",
    standard_cluster="complex_syllable_start complex_syllable_tail",
    broken_cluster="R? (complex_syllable_tail | number_joiner_terminated_cluster_tail | numeral_cluster_tail | symbol_cluster_tail)",
    number_joiner_terminated_cluster="N number_joiner_terminated_cluster_tail",
    numeral_cluster="N numeral_cluster_tail?",
    symbol_cluster="(S | GB) symbol_cluster_tail?",
    hieroglyph_cluster="SB+ | SB* G SE* (J SE* (G SE*)?)*",
    independent_cluster="O",
)
use_categories = set([x[2] for x in database["USECategory.txt"]["data"]])


class USEShaper(SyllabicShaper):
    basic_features = ["rkrf", "abvf", "blwf", "half", "pstf", "vatu", "cjct"]
    topographical_features = ["isol", "init", "medi", "fina"]
    other_features = ["abvs", "blws", "haln", "pres", "psts"]
    repha = "R"
    syllable_machine = make_syllable_machine(states, additional_categories=use_categories)
    syllable_types = [
        "independent_cluster",
        "virama_terminated_cluster",
        "sakot_terminated_cluster",
        "standard_cluster",
        "number_joiner_terminated_cluster",
        "numeral_cluster",
        "symbol_cluster",
        "hieroglyph_cluster",
        "broken_cluster",
        "other",
    ]

    def collect_features(self, shaper):
        shaper.add_pause(self.setup_syllables)
        # Default glyph pre-processing
        shaper.add_features("locl", "ccmp", "nukt", "akhn")
        shaper.add_pause(self.clear_substitution)
        shaper.add_features("rphf")
        shaper.add_pause(self.record_rphf_use)
        shaper.add_pause(self.clear_substitution)
        shaper.add_features("pref")
        shaper.add_pause(self.record_pref_use)

        shaper.add_features(*self.basic_features)
        shaper.add_pause(self.initial_reordering)
        # shaper.add_pause(self.clear_syllables)
        shaper.add_features(*self.topographical_features)
        shaper.add_pause()
        shaper.add_features(*self.other_features)

    def assign_category(self, item):
        item.syllabic_category = ucd_data(item.codepoint).get("USE_Category", "X")
        # Separate positional categories are not used, it's all in the syllabic_category
        item.positional_category = "x"

    def cat(self, i):
        return self.buffer.items[i].syllabic_category

    def setup_syllables(self, shaper):
        super().setup_syllables(shaper)
        self.setup_rphf_mask()
        self.setup_topographical_masks()

    def setup_rphf_mask(self):
        for index, syll_type, start, end in self.iterate_syllables():
            if self.cat(start) == "R":
                limit = 1
            else:
                limit = min(3, end - start)
            for i in range(start, end):
                self.buffer.items[i].feature_masks["rphf"] = i > start + limit

    def clear_substitution(self, shaper):
        for i in self.buffer.items:
            i.substituted = False

    def record_pref_use(self, shaper):
        for i in self.buffer.items:
            if i.substituted:
                i.syllabic_category = "VPre"

    def record_rphf_use(self, shaper):
        for i in self.buffer.items:
            if i.substituted:
                i.syllabic_category = "R"

    def is_post_base(self, i):
        return self.cat(i) in [
            "FAbv",
            "FBlw",
            "FPst",
            "MAbv",
            "MBlw",
            "MPst",
            "MPre",
            "VAbv",
            "VBlw",
            "VPst",
            "VPre",
            "VMAbv",
            "VMBlw",
            "VMPst",
            "VMPre",
        ]

    def is_halant(self, i):
        return self.cat(i) in ["H", "HVM"] and not self.buffer.items[i].ligated

    def reorder_syllable(self, start, end):
        if self.cat(start) == "R":
            for i in range(start + 1, end):
                post_base = self.is_post_base(i) or self.is_halant(i)
                if post_base or i == end - 1:
                    if post_base:
                        i -= 1
                    self.buffer.merge_clusters(start, i + 1)
                    self.buffer.move_item(src=start, dest=i)
                    break
        j = start
        for i in range(start, end):
            if self.is_halant(i):
                j = i + 1
            if self.cat(i) in ["VPre", "VMPre"] and j < i:  # XXX Only first ligcomp
                self.buffer.merge_clusters(j, i + 1)
                self.buffer.move_item(src=i, dest=j)

    initial_reordering_syllable = {
        "virama_terminated_cluster": reorder_syllable,
        "sakot_terminated_cluster": reorder_syllable,
        "standard_cluster": reorder_syllable,
        "broken_cluster": reorder_syllable,
    }

    def setup_topographical_masks(self):
        # Right now we don't actually send joining scripts through the USE
        pass

    def normalize_unicode_buffer(self):
        unicodes = [item.codepoint for item in self.buffer.items]
        newunicodes = []
        for cp in unicodes:
            newunicodes.extend([ord(x) for x in unicodedata.normalize("NFD", chr(cp))])
        # Now recompose
        newstring = ""
        ix = 0
        while ix < len(newunicodes):
            a = newunicodes[ix]
            if ix + 1 == len(newunicodes):
                newstring = newstring + chr(a)
                break

            b = newunicodes[ix + 1]
            s = chr(a) + chr(b)
            composed = unicodedata.normalize("NFC", s)
            if ucd_data(a)["General_Category"][0] == "M":
                newstring = newstring + chr(a)
                ix = ix + 1
                continue
            elif composed != unicodedata.normalize("NFD", s):
                assert len(s) == 1
                newunicodes[ix] = ord(x)
                del newunicodes[ix + 1]
                continue
            else:
                newstring = newstring + chr(a)
                ix = ix + 1

        self.buffer.store_unicode(newstring)



"""ttLib.ValueRecord: Converting ValueRecord rules to TrueType."""

from fontTools.ttLib.tables.otBase import ValueRecord as OTLValueRecord
from fontTools.feaLib.variableScalar import VariableScalar
from fontTools.varLib.builder import buildVarDevTable


def toOTValueRecord(self, ff, pairPosContext=False):
    """Converts the ValueRecord to an ``OTLValueRecord`` object. If the
    value record contains any variable scalars, they are saved to the
    GDEF variation store."""
    otl_value = OTLValueRecord()
    if pairPosContext and not self:
        self.XAdvance = 0
    for item in ["xPlacement", "yPlacement", "xAdvance", "yAdvance"]:
        itemvalue = getattr(self, item)
        if not itemvalue:
            continue
        item = item[0].upper() + item[1:]
        if isinstance(itemvalue, VariableScalar):
            itemvalue, index = itemvalue.add_to_variation_store(ff.varstorebuilder)
            if index != 0xFFFF:
                setattr(otl_value, item[0:4]+"Device", buildVarDevTable(index))
        setattr(otl_value, item, itemvalue)
    return otl_value



from fontFeatures import Substitution
from fontFeatures.feaLib import FeaParser

import pytest

import re

def assertSufficientlyEqual(s1, s2):
    def alltrim(a):
        a = re.sub("lookupflag 0;", "", a)
        a = re.sub("#.*", "", a)
        a = re.sub("\\s+", " ", a)
        a = re.sub("table GDEF.*GDEF;", "", a)
        a = re.sub(" ; ", " ", a)
        return a.strip()

    assert alltrim(s1) == alltrim(s2)

@pytest.mark.parametrize("s", [
    # pytest.param("lookup dummy { } dummy;", id="single_dummy"),
    # pytest.param("lookup dummy1 { } dummy1; lookup dummy2 { } dummy2;", id="double_dummy"),
    pytest.param("lookup dummy { sub a by b; } dummy;", id="single_subst"),
    pytest.param("lookup dummy { sub x a' y by b; } dummy;", id="single_marked_subst"),
    pytest.param("lookup dummy { sub [a b] by [d e]; } dummy;", id="single_group_subst"),
    pytest.param("lookup dummy { sub [a b]' y by [d e]; } dummy;", id="single_marked_group_subst"),
    pytest.param("lookup dummy { sub one two [a b]' y by [d e]; } dummy;", id="single_prepost_1"),
    pytest.param("lookup dummy { sub [one two] [a b]' y by [d e]; } dummy;", id="single_prepost_2"),
    pytest.param("lookup dummy { sub f_f_i by f f i; } dummy;", id="decomposition"),
    pytest.param("lookup dummy { sub f_f_i' x by f f i; } dummy;", id="decomposition_marked"),
    pytest.param("lookup dummy { sub [a b] f_f_i' x by f f i; } dummy;", id="decomposition_context"),
    pytest.param("lookup dummy { sub ampersand from [ampersand.1 ampersand.2 ampersand.3]; } dummy;", id="alternate"),
    pytest.param("lookup dummy { sub x ampersand' from [ampersand.1 ampersand.2 ampersand.3]; } dummy;", id="contextual_alternate"),
    pytest.param("lookup dummy { sub [one one.oldstyle] [slash fraction] [two two.oldstyle] by onehalf; } dummy;",id="fraction"),
    pytest.param("lookup dummy { sub [o u] f' f' by f_f; } dummy;",id="marked_ligature"),
    pytest.param("lookup dummy { sub [e e.begin]' t' c by ampersand; } dummy;",id="contextual"),
    pytest.param("lookup dummy { sub a by b; } dummy; lookup chain { sub a' lookup dummy; } chain; ",id="chain"),
    pytest.param("lookup dummy { sub a by b; } dummy; lookup chain { sub x [one two] a' lookup dummy b' lookup dummy; } chain; ",id="chain2"),
    pytest.param("lookup dummy { rsub a b' c by d; } dummy; ",id="rsub"),
    pytest.param("lookup dummy { ignore sub a d' d; } dummy; ",id="ignore"),
    pytest.param("lookup a { sub a by b; sub c by d; } a; feature calt { lookup a; } calt;",id="feature"),
    pytest.param("lookup dummy { pos one <-80 0 -160 0>; } dummy;",id="pos_one"),
    pytest.param("lookup dummy { pos [one two] <-80 0 -160 0>; } dummy;",id="pos_one_group"),
    pytest.param("lookup dummy { pos a [one two]' <-80 0 -160 0>; } dummy;",id="pos_one_contexual"),
    pytest.param("lookup dummy { pos s f' 10 t; } dummy;",id="pos_one_contextual_alt_format"),
    pytest.param("lookup dummy { pos T -60 a <-40 0 -40 0>; } dummy;",id="pos_two"),
    pytest.param("lookup dummy { pos [X Y] -60 a <-40 0 -40 0>; } dummy;",id="pos_two_group"),
    pytest.param("lookup dummy { pos T a -100; } dummy;",id="pos_two_one_null"),
    pytest.param("lookup dummy { pos cursive meem.medial <anchor 500 20> <anchor 0 -20>; } dummy;",id="pos_cursive1"),
    pytest.param("lookup dummy { pos cursive meem.init <anchor NULL> <anchor 0 -20>; } dummy;",id="pos_cursive2"),
    pytest.param(
        """markClass [acute grave] <anchor 150 -10> @TOP_MARKS;
  lookup dummy {
  pos base [a e o u] <anchor 250 450> mark @TOP_MARKS;
  } dummy;
  """, id="mark_attachment"
        ),
    pytest.param("lookup dummy { pos X [A B]' -40 B' -40 A' -40 Y; } dummy;", id="pos_contextual"),
    # pytest.param("lookup dummy { sub a b by c; } dummy; feature calt { lookup dummy; } calt;", id="lookup_reference"),
    pytest.param("lookup dummy { lookupflag IgnoreMarks; sub a b by c; } dummy;", id="lookup_flag"),
    # pytest.param("feature calt { lookupflag IgnoreMarks; sub a b by c; } calt;", id="flag_in_feature"),
    # pytest.param("feature calt { sub x by y; lookupflag IgnoreMarks; sub a b by c; } calt;", id="switch_flags"),
])
def test_round_trip(s):
    parser = FeaParser(s)
    parser.parse()
    assertSufficientlyEqual(parser.ff.asFea(), s)

# Mark to mark
# Mark to lig
# Ignore pos
# Languages



from fontFeatures import Attachment, FontFeatures
from lxml import etree
import re

import unittest

def assertSufficientlyEqual(s1, s2):
    def alltrim(a):
        a = re.sub("lookupflag 0;", "", a)
        a = re.sub("#.*", "", a)
        a = re.sub("\\s+", " ", a)
        a = re.sub("table GDEF.*GDEF;", "", a)
        a = re.sub(" ; ", " ", a)
        return a.strip()

    assert alltrim(s1) == alltrim(s2)


class TestAnchors(unittest.TestCase):
    def roundTrip(self, thing):
      rt = thing.__class__.fromXML(thing.toXML())
      self.assertEqual(rt.asFea(), thing.asFea())

    def test_markbase(self):
        s = Attachment("top", "top_", {"A": (679, 1600)}, {"acutecomb": (-570, 1290)})
        assertSufficientlyEqual(s.asFea(), "    pos base A <anchor 679 1600> mark @top;\n")
        self.assertEqual(s.involved_glyphs, set(["A", "acutecomb"]))
        self.assertEqual(etree.tostring(s.toXML()), '<attachment basename="top" markname="top_"><base name="A" anchorX="679" anchorY="1600"/><mark name="acutecomb" anchorX="-570" anchorY="1290"/></attachment>'.encode("utf-8"))
        self.roundTrip(s)



from fontFeatures.fontDameLib import unparse
import re


def assertSufficientlyEqual(s1, s2):
    def alltrim(a):
        a = re.sub("lookupflag 0;", "", a)
        a = re.sub("#.*", "", a)
        a = re.sub("\\s+", " ", a)
        a = re.sub("table GDEF.*GDEF;", "", a)
        a = re.sub(" ; ", " ", a)
        return a.strip()

    assert alltrim(s1) == alltrim(s2)


def test_noto_sharada():
    ff = unparse("tests/data/Noto Sans Sharada GSUB.txt")
    assert "abvs" in ff.features
    assert "akhn" in ff.features
    assert "blws" in ff.features
    assert len(ff.routines) == 12
    r1 = ff.features["abvs"][0].routine.asFea()
    assertSufficientlyEqual(r1, """
    sub Jihvamuliya by Jihvamuliya.ns;
    sub Upadhmaniya by Upadhmaniya.ns;
""")
    r2 = ff.features["blws"][1].routine.asFea()
    assertSufficientlyEqual(r2, """
sub [KKa NgKa RKa RGa JNya TtKa NnSha Sa]' [U UU]' lookup lookup_9;
sub [KKa NgKa RKa RGa JNya TtKa NnSha Sa]' [UU.alt U.alt]' lookup lookup_9;
sub [KTa RTa U UU]' [U UU]' lookup lookup_11;
sub [KTa RTa U UU]' [UU.alt U.alt]' lookup lookup_11;
sub [KKa NgKa RKa RGa JNya TtKa NnSha Sa]' [KTa RTa U UU]' lookup lookup_10;
""")




"""ttLib: Interfacing with TrueType fonts.

This package contains routines for converting between fontTools objects
(representing TrueType/OpenType fonts) and fontFeatures. This particular
module is mainly concerned with getting information out of binary OTF/TTF
fonts and into fontFeatures."""
from collections import OrderedDict

from .GDEFUnparser import GDEFUnparser
from .GSUBUnparser import GSUBUnparser
from .GPOSUnparser import GPOSUnparser


def unparseLanguageSystems(tables):
    """Build a set of script / language pairs from a GSUB/GPOS table.

    Args:
        tables: A list of ``fontTools.ttLib.tables.G_S_U_B_.table_G_S_U_B_`` /
            ``fontTools.ttLib.tables.G_P_O_S_.table_G_P_O_S_`` objects.

    Returns an ordered dictionary whose keys are four-character script tags and
      whose values are a set of four-character language tags.
    """
    scripts = OrderedDict()
    for table in tables:
        if not table.table.ScriptList:
            continue
        for scriptRecord in table.table.ScriptList.ScriptRecord:
            scriptTag = scriptRecord.ScriptTag
            languages = scripts.get(scriptTag, [])
            script = scriptRecord.Script
            items = []
            if script.DefaultLangSys is not None:
                items.append(("dflt", script.DefaultLangSys))
            items += [(l.LangSysTag, l.LangSys) for l in script.LangSysRecord]
            languages = set([i[0] for i in items])

            if languages and not scriptTag in scripts:
                scripts[scriptTag] = languages

    return scripts


def unparse(font, do_gdef=False, doLookups=True, config={}):
    """Convert a binary OpenType font into a fontFeatures object

    Args:
        font: A ``TTFont`` object.
        do_gdef: Boolean. Whether the GDEF table should also be read.
        doLookups: Whether the lookups should be read, or just the script/language
            information and top-level features.
        config: A dictionary of glyph class and routine names.
    """
    gsub_gpos = [font[tableTag] for tableTag in ("GSUB", "GPOS") if tableTag in font]
    from fontFeatures import FontFeatures
    ff = FontFeatures()

    languageSystems = unparseLanguageSystems(gsub_gpos)

    if "GSUB" in font:
        GSUBUnparser(
            font["GSUB"], ff, languageSystems, font=font, config=config
        ).unparse(doLookups=doLookups)

    if "GPOS" in font:
        GPOSUnparser(
            font["GPOS"], ff, languageSystems, font=font, config=config
        ).unparse(doLookups=doLookups)

    if "GDEF" in font and do_gdef:
        GDEFUnparser(font["GDEF"], ff).unparse()

    return ff



import copy

def shaper_inputs(self):
    return self.input


def _do_apply(self, buf, ix, namedclasses={}):
    from fontFeatures.shaperLib.Rule import _expand_slot

    coverage = buf[ix : ix + len(self.input)]
    newstuff = []
    # Handle single subst first
    if len(self.input) == 1 and len(self.replacement) == 1:
        replacements = _expand_slot(self.replacement[0], namedclasses)
        if len(replacements) == 1:
            buf[ix].glyph = replacements[0]
        else:
            inputs = _expand_slot(self.input[0], namedclasses)
            buf[ix].glyph = replacements[inputs.index(buf[ix].glyph)]
        buf[ix].prep_glyph(buf.font)
        return

    delta = len(self.replacement) - 1
    for g in self.replacement:
        new_glyph = copy.copy(buf[ix])
        new_glyph.glyph = g[0]
        new_glyph.prep_glyph(buf.font)
        new_glyph.substituted = True
        if len(self.replacement) > 1 and len(coverage) == 1:
            new_glyph.multiplied = True
        if len(self.replacement) == 1 and len(coverage) > 1:
            new_glyph.ligated = True
        newstuff.append(new_glyph)
    buf[ix : ix + len(self.input)] = newstuff
    if delta > 0:
        return delta



"""ttLib.Attachment: Converting Attachment rules to TrueType."""

from glyphtools import categorize_glyph


def lookup_type(self):
    """Mixin to determine the GPOS lookup type of a fontFeatures.Attachment object

    Returns: integer GPOS lookup type."""
    if self.is_cursive:
        return 3
    # Terrible hacks
    firstbase = list(self.bases.keys())[0]
    if hasattr(self, "fontfeatures") and self.fontfeatures.glyphclasses.get(firstbase) == "mark":
        return 6
    if self.font and categorize_glyph(self.font,firstbase)[0] == "mark":
        return 6
    return 4



from fontTools.otlLib.builder import ClassDefBuilder
from fontFeatures import Substitution, Positioning, Chaining
import logging
import copy


class MoveLongCoverageToClassDefinition:
    level = 1

    def gensym(self, ff):
        if not "index" in ff.scratch:
            ff.scratch["index"] = 0
        ff.scratch["index"] = ff.scratch["index"] + 1
        return str(ff.scratch["index"])

    def replaceLongWithClasses(self, i, ff):
        for ix, gc in enumerate(i):
            if len(gc) > 5:
                classname = ff.getNamedClassFor(gc, "class" + self.gensym(ff))
                i[ix] = ["@" + classname]

    def apply(self, routine, ff):
        for rule in routine.rules:
            if isinstance(rule, Substitution):
                self.replaceLongWithClasses(rule.input, ff)
                self.replaceLongWithClasses(rule.precontext, ff)
                self.replaceLongWithClasses(rule.postcontext, ff)
                self.replaceLongWithClasses(rule.replacement, ff)
            if isinstance(rule, Positioning):
                self.replaceLongWithClasses(rule.glyphs, ff)
                self.replaceLongWithClasses(rule.precontext, ff)
                self.replaceLongWithClasses(rule.postcontext, ff)

        return []


class MergeMultipleSingleSubstitutions:
    level = 1

    def apply(self, routine, ff):

        _is_single_sub = (
            lambda rule: isinstance(rule, Substitution)
            and len(rule.input) == 1
            and len(rule.replacement) == 1
        )

        if len(routine.rules) < 2:
            return
        new_rules = [routine.rules[0]]
        for r in routine.rules[1:]:
            previous = new_rules[-1]
            if (
                _is_single_sub(r)
                and _is_single_sub(previous)
                and r.precontext == previous.precontext
                and r.postcontext == previous.postcontext
            ):
                new_rules.pop()
                new_rules.append(self.merge_two(previous, r))
            else:
                new_rules.append(r)
        routine.rules = new_rules
        # It's possible to get clever later with non-adjacent substitutions
        # if there's nothing in the middle that could affect the output

    def merge_two(self, first, second):
        assert len(first.input) == 1 and len(second.input) == 1
        assert len(first.replacement) == 1 and len(second.replacement) == 1
        assert first.precontext == second.precontext
        assert first.postcontext == second.postcontext
        firstmapping = {l: r for l, r in zip(first.input[0], first.replacement[0])}
        secondmapping = {l: r for l, r in zip(second.input[0], second.replacement[0])}
        for l, r in firstmapping.items():
            if r in secondmapping:
                firstmapping[l] = secondmapping[r]
        for l, r in secondmapping.items():
            if not (l in firstmapping):
                firstmapping[l] = r
        logger = logging.getLogger("fontFeatures")
        logger.info("Merging two adjacent single subs")
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(first.asFea())
            logger.debug(second.asFea())
        address = first.address or second.address
        return Substitution(
            [list(firstmapping.keys())],
            [list(firstmapping.values())],
            address=address,
            precontext=first.precontext,
            postcontext=first.postcontext,
        )

class EnsureFormat2Chaining:
    level = 1

    def apply(self, routine, ff):
        rules = routine.rules
        if not all(isinstance(rule,Chaining) for rule in routine.rules):
            return

        while True:
            if self._apply(rules):
                break

    def _apply(self, rules):
        any_failing = False
        logger = logging.getLogger("fontFeatures")
        logger.warn("Applying format 2 optimization round")

        for which in ["input", "precontext", "postcontext"]:
            classdefbuilder = ClassDefBuilder(useClass0=False)
            for rule in rules:
                for slot in getattr(rule, which):
                    if not classdefbuilder.canAdd(set(slot)):
                        logger.warn("Mitigating. Rule count before=%i", len(rules))
                        self.mitigate(rules, which, classdefbuilder.classes(), set(slot))
                        logger.warn("Mitigating. Rule count after=%i", len(rules))
                        return False
                    classdefbuilder.add(set(slot))
        return True

    def mitigate(self, rules, which, rule_classes, failing_slot):
        # What kind of fail is this?
        problem = None
        for klass in rule_classes[1:]:
            if not set(klass).isdisjoint(failing_slot):
                problem = klass
                break
        # Split both into two and try again
        additional_rules = []
        intersection = set(problem) & failing_slot
        assert intersection
        for rule in rules:
            for ix, slot in enumerate(getattr(rule, which)):
                if (set(slot) == set(problem) or set(slot) == failing_slot) and set(slot) != intersection:
                    new_rule = Chaining(
                        copy.deepcopy(rule.input),
                        precontext=copy.deepcopy(rule.precontext),
                        postcontext=copy.deepcopy(rule.postcontext),
                        lookups=rule.lookups[:]
                    )
                    setattr(new_rule, which, getattr(new_rule, which)[:])
                    setattr(rule, which, getattr(rule, which)[:])
                    getattr(rule, which)[ix] = list(set(slot) - intersection)
                    getattr(new_rule, which)[ix] = list(intersection)
                    additional_rules.append(new_rule)
        rules.extend(additional_rules)




optimizations = [
    MergeMultipleSingleSubstitutions,
    EnsureFormat2Chaining,
    # MoveLongCoverageToClassDefinition,  # Runs last
]



from youseedee import ucd_data
from .BaseShaper import BaseShaper
import re
from fontFeatures.shaperLib.Buffer import BufferItem
from fontFeatures.shaperLib.VowelConstraints import preprocess_text_vowel_constraints
from .IndicShaperData import script_config, syllabic_category_map, syllable_machine_indic, IndicPositionalCategory2IndicPosition, IndicPosition, reassign_category_and_position_indic
from .SyllabicShaper import SyllabicShaper
import unicodedata

DOTTED_CIRCLE = 0x25CC

class IndicShaper(SyllabicShaper):

    syllable_machine = syllable_machine_indic
    syllable_types = ["consonant_syllable", "vowel_syllable", "standalone_cluster","symbol_cluster","broken_cluster","other"]

    @property
    def config(self):
        return script_config.get(self.buffer.script, script_config["Invalid"])

    def override_features(self, shaper):
        shaper.disable_feature("liga")

    def consonant_position_from_face(self, consonant):
        virama = self.config["virama"]
        consonant_item = BufferItem.new_unicode(consonant)
        virama_item = BufferItem.new_unicode(virama)
        consonant_item.map_to_glyph(self.buffer.font)
        virama_item.map_to_glyph(self.buffer.font)

        if self.would_substitute("blwf", [virama_item, consonant_item]):
            return IndicPosition.BELOW_C
        if self.would_substitute("blwf", [consonant_item, virama_item]):
            return IndicPosition.BELOW_C
        if self.would_substitute("vatu", [virama_item, consonant_item]):
            return IndicPosition.BELOW_C
        if self.would_substitute("vatu", [consonant_item, virama_item]):
            return IndicPosition.BELOW_C

        if self.would_substitute("pstf", [virama_item, consonant_item]):
            return IndicPosition.POST_C
        if self.would_substitute("pstf", [consonant_item, virama_item]):
            return IndicPosition.POST_C

        if self.would_substitute("pref", [virama_item, consonant_item]):
            return IndicPosition.POST_C
        if self.would_substitute("pref", [consonant_item, virama_item]):
            return IndicPosition.POST_C
        return IndicPosition.BASE_C

    def initial_reordering_pre(self):
        if self.config["base_pos"] == "last": # Not Sinhala
            for item in self.buffer.items:
                if item.syllabic_position == IndicPosition.BASE_C:
                    item.syllabic_position = self.consonant_position_from_face(item.codepoint)
                    pass

    def reassign_category(self, item):
        reassign_category_and_position_indic(item)

    def initial_reordering_consonant_syllable(self, start, end):
        def cat(i):
            return self.buffer.items[i].syllabic_category
        def pos(i):
            return self.buffer.items[i].syllabic_position
        def swap(a,b):
            self.buffer.items[b], self.buffer.items[a] = self.buffer.items[a], self.buffer.items[b]
        def is_joiner(n):
            return cat(n) == "ZWJ" or cat(n) == "ZWNJ"
        def is_consonant(n):
            isc = cat(n)
            is_medial = isc == "CM"
            return isc in ["C", "CS", "Ra", "V", "PLACEHOLDER", "DOTTEDCIRCLE"] or is_medial

        if self.buffer.script == "Kannada" and start + 3 <= end and cat(start) == "Ra" and cat(start+1) == "H" and cat(start+2) == "ZWJ":
            swap(start+1, start+2)

        syllable_index = self.buffer.items[start].syllable_index

        base = end
        has_reph = False

        limit = start
        if "rphf" in self.plan.fontfeatures.features and start + 3 <= end \
            and ( \
                (self.config["reph_mode"] == "implicit" and not is_joiner(start+2)) \
                or (self.config["reph_mode"] == "explicit" and cat(start+2) == "ZWJ") \
            ):
            if self.would_substitute("rphf", self.buffer.items[start:start+2]) \
                or self.would_substitute("rphf", self.buffer.items[start:start+3]):
                limit = limit + 2
                while limit < end and is_joiner(limit):
                    limit = limit + 1
                base = start
                has_reph = True
        elif self.config["reph_mode"] == "log_repha" and cat(start) == "Repha":
            limit = limit + 1
            while limit < end and is_joiner(limit):
                limit = limit + 1
            base = start
            has_reph = True

        if self.config["base_pos"] == "last":
            i = end
            seen_below = False
            while True:
                i = i -1
                if is_consonant(i):
                    if pos(i) != IndicPosition.BELOW_C and \
                        (pos(i) != IndicPosition.POST_C or seen_below):
                        base = i
                        break
                    if pos(i) == IndicPosition.BELOW_C:
                        seen_below = True
                    base = i
                else:
                    if start < i and cat(i) == "ZWJ" and cat(i-1) == "H":
                        break
                if i <= limit:
                    break
        elif self.config["base_pos"] == "last_sinhala":
            if not has_reph:
                base = limit
            for i in range(limit, end):
                if is_consonant(i):
                    if limit < i and cat(i-1) == "ZWJ":
                        break
                    else:
                        base = i
            for i in range(base+1, end):
                if is_consonant(i):
                    self.buffer.items[i].syllabic_position = IndicPosition.BELOW_C

        if has_reph and base == start and limit - base <= 2:
            has_reph = False

        self.plan.msg("Base consonant for syllable %i is %s" % (syllable_index, self.buffer.items[base].glyph))
        for i in range(start, base):
            self.buffer.items[i].syllabic_position = min(IndicPosition.PRE_C, pos(i))
        if base < end:
            self.buffer.items[i].syllabic_position = IndicPosition.BASE_C

        # Mark final consonants
        for i in range(base+1, end):
            if cat(i) == "M":
                for j in range(i, end):
                    if is_consonant(j):
                        self.buffer.items[j].syllabic_position = IndicPosition.FINAL_C
                        break
                break

        if has_reph:
            self.buffer.items[start].syllabic_category = IndicPosition.RA_TO_BECOME_REPH

        if self.config["old_spec"]:
            disallow_double_halants = self.buffer.script == "Kannada"
            for i in range(base+1, end):
                if cat(i) == "H":
                    j = end - 1
                    while j > i:
                        if is_consonant(j) or (disallow_double_halants and cat(j) == "H"):
                            break
                        j = j - 1
                    if cat(j) != "H" and j > i:
                        self.buffer.items.insert(j, self.buffer.items.pop(i))
                        self.plan.msg("Moved double halant", self.buffer)
                    break

        last_pos = IndicPosition.START
        for i in range(start, end):
            if cat(i) in ["ZWJ", "ZWNJ", "N", "RS", "CM", "H"]:
                self.buffer.items[i].syllabic_position = last_pos
                if cat(i) == "H" and pos(i) == IndicPosition.PRE_M:
                    for j in range(i,start,-1):
                        if pos(j-1) != IndicPosition.PRE_M:
                            self.buffer.items[i].syllabic_position = pos(j-1)
                            break
            elif pos(i) != IndicPosition.SMVD:
                last_pos = pos(i)

        last = base
        for i in range(base+1, end):
            if is_consonant(i):
                for j in range(last+1, i):
                    if pos(j) < IndicPosition.SMVD:
                        self.buffer.items[j].syllabic_position = pos(i)
                last = i
            elif cat(i) == "M":
                last = i

        # As with Harfbuzz, temporarily abuse syllable index
        for i in range(start, end):
            self.buffer.items[i].syllable_index = start - i

        # REORDER
        self.buffer.items[start:end] = sorted(self.buffer.items[start:end], key=lambda x:x.syllabic_position)

        base = end
        for i in range(start, end):
            if pos(i) == IndicPosition.BASE_C:
                base = i
                break

        if self.config["old_spec"] or end - start > 127:
            # Merge clusters
            pass
        else:
            for i in range(base, end):
                if self.buffer.items[i].syllable_index != 255:
                    max_i = i
                    j = start + self.buffer.items[i].syllable_index
                    while j != i:
                        max_i = max(max_i, j)
                        next_i = start + self.buffer.items[j].syllable_index
                        self.buffer.items[j].syllable_index = 255
                        j = next_i
                    if i != max_i:
                        # Merge clusters
                        pass

        for i in range(start, end):
            self.buffer.items[i].syllable_index = syllable_index
        self.plan.msg("After initial reordering", self.buffer)

        # Set up masks now. Note that these masks have the opposite
        # value to Harfbuzz - i.e. False means "not masked"
        rphf_mask = False
        for i in range(start, end):
            if pos(i) != IndicPosition.RA_TO_BECOME_REPH:
                rphf_mask = True
            self.buffer.items[i].feature_masks["rphf"] = rphf_mask
            self.buffer.items[i].feature_masks["half"] = i > base
            if not self.config["old_spec"] and self.config["blwf_mode"] == "pre_and_post":
                self.buffer.items[i].feature_masks["blwf"] = i > base
            self.buffer.items[i].feature_masks["blwf"] = i < base
            self.buffer.items[i].feature_masks["abvf"] = i < base
            self.buffer.items[i].feature_masks["pstf"] = i < base

        # We are not supporting old spec eyelash ra

        # pref substitutes pairwise
        pref_len = 2
        i = base + 1
        for j in range(0,i):
            self.buffer.items[j].feature_masks["pref"] = True
        while i < end-pref_len:
            if self.would_substitute("pref", [self.buffer.items[i], self.buffer.items[i+1]]):
                self.buffer.items[i].feature_masks["pref"] = False
                self.buffer.items[i+1].feature_masks["pref"] = False
                i = i + 2
            else:
                self.buffer.items[i].feature_masks["pref"] = True
                i = i + 1

        # ZWJ/ZWNJ
        for i in range(start+1, end):
            if cat(i) in ["ZWJ", "ZWNJ"]:
                non_joiner = cat(i) == "ZWNJ"
                j = i
                while True:
                    j = j - 1
                    if non_joiner:
                        self.buffer.items[j].feature_masks["half"] = True
                    if not (j > start and not is_consonant(j)):
                        break


    initial_reordering_syllable = {
        "standalone_cluster": initial_reordering_consonant_syllable,
        "consonant_syllable": initial_reordering_consonant_syllable
    }


    def final_reordering_syllable(self, start, end):
        def cat(i):
            return self.buffer.items[i].syllabic_category
        def pos(i):
            return self.buffer.items[i].syllabic_position
        def swap(a,b):
            self.buffer.items[b], self.buffer.items[a] = self.buffer.items[a], self.buffer.items[b]
        def is_joiner(n):
            return cat(n) == "ZWJ" or cat(n) == "ZWNJ"
        def is_halant(n):
            return cat(n) == "H"
        def is_consonant(n):
            isc = cat(n)
            is_medial = isc == "CM"
            return isc in ["C", "CS", "Ra", "V", "PLACEHOLDER", "DOTTEDCIRCLE"] or is_medial

        virama = self.config["virama"]
        virama_item = BufferItem.new_unicode(virama)
        virama_item.map_to_glyph(self.buffer.font)
        if virama_item.glyph != ".notdef":
            for i in range(start, end):
                if self.buffer.items[i].glyph == virama_item.glyph \
                    and self.buffer.items[i].ligated \
                    and self.buffer.items[i].multiplied:
                    self.buffer.items[i].syllabic_category = "H"
                    self.buffer.items[i].ligated = False
                    self.buffer.items[i].multiplied = False
        try_pref = any(["pref" in item.feature_masks and item.feature_masks["pref"] == False for item in self.buffer.items])
        base = start
        while base < end:
            if pos(base) >= IndicPosition.BASE_C:
                if try_pref and base + 1 < end:
                    for i in range(base+1, end):
                        item = self.buffer.items[i]
                        if not item.feature_masks.get("pref",True):
                            if not (item.substituted and (item.ligated and not item.multiplied)):
                                base = i
                                while base < end and is_halant(base):
                                    base = base + 1
                                self.buffer.items[base].syllabic_positional_category = IndicPosition.BASE_C
                                try_pref = false
                            break
                if self.buffer.script == "Malayalam":
                    i = base + 1
                    while i < end:
                        while i < end and is_joiner(i):
                            i = i + 1
                        if i == end or not is_halant(i):
                            break
                        i = i + 1
                        while i < end and is_joiner(i):
                            i = i + 1
                        if i < end and is_consonant(i) and pos(i) == IndicPosition.BELOW_C:
                            base = i
                            self.buffer.items[base].syllabic_positional_category = IndicPosition.BASE_C
                        i = i + 1
                if start < base and pos(base) > IndicPosition.BASE_C:
                    base = base - 1
                break
            base = base + 1
        if base == end and start < base and cat(base-i) == "ZWJ":
            base = base - 1
        if base < end:
            while start < base and cat(base) in ["N","H"]:
                base = base - 1

        # Reorder matras
        if start + 1 < end and start < base:
            new_pos = base -1
            if base == end:
                new_pos = base - 2
            # XXX

        for i in range(start,end):
            self.buffer.items[i].feature_masks["init"] = True
        if pos(start) == IndicPosition.PRE_M:
            reverse_map = {k:v for k,v in self.buffer.font.unicode_map.items()}
            if start == 0 or ucd_data(reverse_map.get(self.buffer.items[start-1].glyph, 0))["General_Category"] not in ["Cf", "Cn", "Co", "Cs", "Ll", "Lm", "Lo", "Lt", "Lu", "Mc", "Me", "Mn"]:
                self.buffer.items[start].feature_masks["init"] = False


    def normalize_unicode_buffer(self):
        unicodes = [item.codepoint for item in self.buffer.items]
        newunicodes = []
        for cp in unicodes:
            if cp in [0x0931, 0x09DC, 0x09DD, 0x0B94]:
                newunicodes.append(cp)
            elif cp in [0x0DDA, 0x0DDC, 0x0DDD, 0x0DDE]: # Sinhala split matras
                glyph = BufferItem.new_unicode(cp)
                glyph.map_to_glyph(self.buffer.font)
                if self.would_substitute("pstf", [glyph]):
                    newunicodes.extend([0x0DD9, cp])
                else:
                    newunicodes.append(cp)
            else:
                newunicodes.extend([ord(x) for x in unicodedata.normalize("NFD", chr(cp)) ])
        # Now recompose
        newstring = ""
        ix = 0
        while ix < len(newunicodes):
            a = newunicodes[ix]
            if ix+1 == len(newunicodes):
                newstring = newstring + chr(a)
                break

            b = newunicodes[ix+1]
            s = chr(a) + chr(b)
            composed = unicodedata.normalize("NFC", s)
            if ucd_data(a)["General_Category"][0] == "M":
                newstring = newstring + chr(a)
                ix = ix + 1
                continue
            elif a == 0x9af and b == 0x9bc:
                newstring = newstring + chr(0x9df)
                ix = ix + 2
                continue
            # elif composed != unicodedata.normalize("NFD", s):
            #     assert(len(s) == 1)
            #     newunicodes[ix] = ord(x)
            #     del newunicodes[ix+1]
            #     continue
            else:
                newstring = newstring + chr(a)
                ix =ix + 1

        self.buffer.store_unicode(newstring)




"""Routines for converting RoutineReference objects to and from Adobe FEA."""
import fontTools.feaLib.ast as feaast
from fontFeatures.ttLib.Substitution import lookup_type as sub_lookup_type
from fontFeatures.ttLib.Positioning import lookup_type as pos_lookup_type

def feaPreamble(self, ff):
    return []


def asFeaAST(self, expand=False):
    # if self.routine.usecount == 1:
    #     return self.routine.asFeaAST()
    f = feaast.Block()
    f.statements.append(feaast.LookupReferenceStatement(self.routine.asFeaAST()))
    return f





def put_languages(self, root):
    if hasattr(self, "languages") and self.languages:
        lang_string = ",".join(["/".join(x) for x in self.languages])
        root.attrib["languages"] = lang_string


def put_address(self, root):
    if self.address:
        if isinstance(self.address, str):
            root.attrib["address"] = self.address
        else:
            root.attrib["address"] = str("|".join(self.address))



from fontFeatures import Chaining, Positioning, ValueRecord, Routine, Substitution, RoutineReference, FontFeatures
from lxml import etree


import unittest


class TestChaining(unittest.TestCase):
    def roundTrip(self, thing, dependencies):
        f = FontFeatures()
        f.routines.extend(dependencies)
        rt = thing.__class__.fromXML(thing.toXML())
        f.routines.append(Routine(rules=[rt]))
        f.resolveAllRoutines()
        self.assertEqual(rt.asFea(), thing.asFea())

    def test_simple_pos(self):
        v = ValueRecord(xAdvance=120)
        pos = Positioning(["a"], [v])
        r = Routine(rules=[pos], name="dummy")
        rr = RoutineReference(routine=r)

        c = Chaining([["a"], ["b"]], lookups=[[rr], None])
        self.assertEqual(c.asFea(), "pos a' lookup dummy b';")
        xml = c.toXML()
        self.assertEqual(etree.tostring(xml), '<chaining><lookups><slot><routinereference name="dummy"/></slot><slot><lookup/></slot></lookups><input><slot><glyph>a</glyph></slot><slot><glyph>b</glyph></slot></input></chaining>'.encode("utf-8"))
        self.roundTrip(c, [r])

    def test_simple_sub(self):
        pos = Substitution(["a"], ["b"])
        r = Routine(rules=[pos], name="dummy")
        rr = RoutineReference(routine=r)

        c = Chaining([["a"], ["b"]], lookups=[[rr], None])
        self.assertEqual(c.asFea(), "sub a' lookup dummy b';")

    def test_ignore(self):
        c = Chaining([["a"], ["b"]], lookups=[])
        self.assertEqual(c.asFea(), "ignore sub a b;")

    def test_complex(self):
        pos1 = Substitution(["a"], ["b"])
        pos2 = Substitution(["b"], ["c"])
        r1 = Routine(rules=[pos1], name="dummy1")
        r2 = Routine(rules=[pos2], name="dummy2")
        rr1 = RoutineReference(routine=r1)
        rr2 = RoutineReference(routine=r2)

        c = Chaining([["a"], ["b"]], lookups=[[rr1, rr2], None])
        self.assertEqual(c.asFea(), "sub a' lookup dummy1 lookup dummy2 b';")

    def test_complex_pos(self):
        v = ValueRecord(xAdvance=120)
        pos1 = Positioning(["a"], [v])
        r1 = Routine(rules=[pos1])
        rr1 = RoutineReference(routine=r1)

        c = Chaining([["a"], ["b"]], lookups=[[rr1], None])
        c.feaPreamble(FontFeatures())
        self.assertEqual(c.asFea(), "pos a' lookup ChainedRoutine1 b';")




"""Routines for converting FontFeatures objects to and from XML."""

from lxml import etree


def toXML(self):
    """Serializes a FontFeatures object to a lxml Element object."""
    root = etree.Element("fontfeatures")
    root.append(xml_glyph_classes(self))

    routines = etree.Element("routines")
    for r in self.routines:
        routines.append(r.toXML())
    root.append(routines)

    features = etree.Element("features")
    for f, routines in self.features.items():
        f_el = etree.Element("feature")
        f_el.set("name", f)
        for r in routines:
            f_el.append(r.toXML())
        features.append(f_el)
    root.append(features)

    anchor_el = etree.Element("anchors")
    for glyph, anchors in self.anchors.items():
        glyph_el = etree.Element("glyph")
        glyph_el.set("name", glyph)
        for name, pos in anchors.items():
            a_el = etree.Element("anchor")
            a_el.set("name", name)
            a_el.set("x", str(pos[0]))
            a_el.set("y", str(pos[1]))
            glyph_el.append(a_el)
        anchor_el.append(glyph_el)
    root.append(anchor_el)

    glyphclasses = etree.Element("glyphclasses")
    for glyph, classname in self.glyphclasses.items():
        glyph_el = etree.Element("glyph")
        glyph_el.set("name", glyph)
        glyph_el.set("class", classname)
        glyphclasses.append(glyph_el)
    root.append(glyphclasses)

    return root


def xml_glyph_classes(self):
    root = etree.Element("namedclasses")
    for name, values in self.namedClasses.items():
        gc = etree.Element("class")
        gc.set("name", name)
        for g in values:
            glyph_el = etree.Element("glyph")
            glyph_el.text = g
            gc.append(glyph_el)
        root.append(gc)
    return root


@classmethod
def fromXML(klass, el):
    """Creates a FontFeatures object from a lxml Element object."""
    f = klass()
    from fontFeatures import RoutineReference, Routine

    for part in el:
        if part.tag == "namedclasses":
            for cl_el in part:
                f.namedClasses[cl_el.get("name")] = [g.text for g in cl_el]
        elif part.tag == "routines":
            f.routines = [Routine.fromXML(r) for r in part]
        elif part.tag == "features":
            for feat in part:
                f.features[feat.get("name")] = [
                    RoutineReference.fromXML(x) for x in feat
                ]
        elif part.tag == "anchors":
            for glyph in part:
                f.anchors[glyph.get("name")] = {
                    el.get("name"): (int(el.get("x")), int(el.get("y"))) for el in glyph
                }
        elif part.tag == "glyphclasses":
            for cl_el in part:
                f.glyphclasses[cl_el.get("name")] = cl_el.get("class")

    for refs in f.features.values():
        for ref in refs:
            ref.resolve(f)
    return f



import io
import re
import fontFeatures
from fontTools.feaLib.parser import Parser
import fontTools.feaLib.ast as ast
from warnings import warn


def bad_statement_to_comment(s):
    # We could check through the attributes but that's hard...
    fea = s.asFea()
    if re.search(r"\[\s*\]", fea):  # Empty classes
        warn("Empty class found in: '"+ fea+ "'")
        return ast.Comment("# " + fea)
    return s


class FeaParser:
    """Turns a AFDKO feature file or string into a FontFeatures object.

    Args:
        featurefile: File object or string.
        font: Optionally, a TTFont object.
        glyphNames: Optionally, a list of glyph names in the font
    """
    def __init__(self, featurefile, font=None, glyphNames=None, includeDir=None):

        self.ff = fontFeatures.FontFeatures()
        self.markclasses = {}
        self.currentFeature = None
        self.currentRoutine = None
        self.gensym = 1
        self.glyphmap = ()
        self.currentLanguage = None
        if font and not glyphNames:
            glyphNames = font.getGlyphOrder()
        if isinstance(featurefile, str):
            featurefile = io.StringIO(featurefile)
        self.featurefile = featurefile
        if glyphNames:
            self.parser = Parser(self.featurefile, glyphNames=glyphNames, includeDir=includeDir)
        else:
            self.parser = Parser(self.featurefile, includeDir=includeDir)
        self.parser.ast.ValueRecord = fontFeatures.ValueRecord

    def parse(self):
        """Parse the feature code.

        Returns:
            A ``FontFeatures`` object containing the rules of this file.
        """

        # Borrow glyph classes
        for name, members in self.ff.namedClasses.items():
            glyphclass = ast.GlyphClassDefinition(
                name, ast.GlyphClass([m for m in members])
            )
            self.parser.glyphclasses_.define(name, glyphclass)

        parsetree = self.parser.parse()

        # Return glyph classes
        if len(self.parser.glyphclasses_.scopes_):
            for name, definition in self.parser.glyphclasses_.scopes_[-1].items():
                if isinstance(definition, ast.MarkClass):
                    pass
                    # self.ff.namedClasses[name] = list(definition.glyphs.keys())
                else:
                    self.ff.namedClasses[name] = definition.glyphs.glyphs

        self.features_ = {}
        parsetree.build(self)
        return self.ff

    def _start_routine_if_necessary(self, location):
        if not self.currentRoutine:
            self._start_routine(location, "")

    def _start_routine(self, location, name):
        location = "%s:%i:%i" % (location)
        self._discard_empty_routine()
        self.currentRoutine = fontFeatures.Routine(name=name, address=location)
        if not name:
            self.currentRoutine.name = "unnamed_routine_%i" % self.gensym
            self.gensym = self.gensym + 1
        self.currentRoutineFlag = 0
        if self.currentFeature:
            reference = self.ff.referenceRoutine(self.currentRoutine)
            self.ff.addFeature(self.currentFeature, [reference])
        else:
            self.ff.routines.append(self.currentRoutine)

    def start_lookup_block(self, location, name):
        self._start_routine(location, name)

    def start_feature(self, location, name):
        self.currentFeature = name

    def set_font_revision(self, location, revision):
        pass

    def add_name_record(self, *args):
        pass

    def add_featureName(self, tag):
        # XXX support this
        pass

    def set_script(self, location, script):
        self.currentLanguage = [(script, "*")]

    def set_language(self, location, language, include_default, required):
        self.currentLanguage = [(self.currentLanguage[0][0], language)]

    def add_single_subst(self, location, prefix, suffix, mapping, forceChain):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Substitution(
            input_=[list(mapping.keys())],
            replacement=[list(mapping.values())],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)


    def add_reverse_chain_single_subst(self, location, prefix, suffix, mapping):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Substitution(
            input_=[list(mapping.keys())],
            replacement=[list(mapping.values())],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage,
            reverse=True
        )
        self.currentRoutine.addRule(s)

    def add_multiple_subst(
        self, location, prefix, glyph, suffix, replacements, forceChain
    ):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Substitution(
            input_=[[glyph]],
            replacement=[[g] for g in replacements],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_alternate_subst(self, location, prefix, glyph, suffix, replacement):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Substitution(
            input_=[[glyph]],
            replacement=[replacement],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_ligature_subst(
        self, location, prefix, glyphs, suffix, replacement, forceChain
    ):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Substitution(
            input_=[list(x) for x in glyphs],
            replacement=[[replacement]],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_chain_context_subst(self, location, prefix, glyphs, suffix, lookups):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        # Find named feature
        mylookups = []
        for x in lookups:
            if x:
                mylookups.append([self.ff.routineNamed(y.name) for y in x])
            else:
                mylookups.append(None)
        s = fontFeatures.Chaining(
            input_=[list(x) for x in glyphs],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            lookups=mylookups,
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    add_chain_context_pos = add_chain_context_subst

    def add_single_pos(self, location, prefix, suffix, pos, forceChain):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Positioning(
            glyphs=[p[0] for p in pos],
            valuerecords=[p[1] for p in pos],
            precontext= [[str(g) for g in group] for group in prefix],
            postcontext= [[str(g) for g in group] for group in suffix],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_specific_pair_pos(self, location, glyph1, value1, glyph2, value2):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Positioning(
            glyphs=[[glyph1], [glyph2]], valuerecords=[value1, value2], address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_class_pair_pos(self, location, glyphclass1, value1, glyphclass2, value2):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        s = fontFeatures.Positioning(
            glyphs=[glyphclass1, glyphclass2],
            valuerecords=[value1, value2],
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_cursive_pos(self, location, glyphclass, entryAnchor, exitAnchor):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        basedict, markdict = {}, {}
        if entryAnchor:
            basedict = {g: (entryAnchor.x, entryAnchor.y) for g in glyphclass}
        if exitAnchor:
            markdict = {g: (exitAnchor.x, exitAnchor.y) for g in glyphclass}
        s = fontFeatures.Attachment(
            base_name="cursive_entry",
            mark_name="cursive_exit",
            bases=basedict,
            marks=markdict,
            address=location,
            languages=self.currentLanguage
        )
        self.currentRoutine.addRule(s)

    def add_mark_base_pos(self, location, bases, marks):
        self._start_routine_if_necessary(location)
        location = "%s:%i:%i" % (location)
        for baseanchor, markclass in marks:
            for definition in markclass.definitions:
                markanchor = definition.anchor
                s = fontFeatures.Attachment(
                    base_name=markclass.name,
                    mark_name=markclass.name,
                    bases={g: (baseanchor.x, baseanchor.y) for g in bases},
                    marks={
                        g: (markanchor.x, markanchor.y) for g in markclass.glyphs.keys()
                    },
                    address=location,
                    languages=self.currentLanguage
                )
                s.fontfeatures = self.ff
                self.currentRoutine.addRule(s)

    add_mark_mark_pos = add_mark_base_pos

    def set_lookup_flag(self, location, value, markAttach, markFilter):
        if self.currentRoutine and value == self.currentRoutineFlag:
            return
        # If we're mid-feature, start a new routine here
        if self.currentFeature:
            self.end_lookup_block()
            self._discard_empty_routine()
            self._start_routine(location, None)
        self.currentRoutineFlag = value

    def add_language_system(self, location, script, language):
        pass

    def setElidedFallbackName(self, location, *args):
        pass
    def addDesignAxis(self, location, *args):
        pass
    def addAxisValueRecord(self, location, *args):
        pass
    def add_ligatureCaretByPos_(self, location, *args): # Urgh
        pass

    def add_lookup_call(self, lookup_name):

        routine = self.ff.routineNamed(lookup_name)
        if self.currentFeature:
            self._discard_empty_routine()
            self.ff.addFeature(self.currentFeature, [routine])
        else:
            raise ValueError("Huh?")

    def end_lookup_block(self):
        if self.currentRoutine:
            for rule in self.currentRoutine.rules:
                rule.flags = self.currentRoutineFlag

    def end_feature(self):
        self._discard_empty_routine()
        self.currentFeature = None
        self.currentLanguage = None
        if self.currentRoutine:
            for rule in self.currentRoutine.rules:
                rule.flags = self.currentRoutineFlag
            self.currentRoutine = None

    def _discard_empty_routine(self):
        if not self.currentFeature:
            return

        # If there genuinely is an empty routine, we should keep it, no?
        # The only problem is that for some reason, we're generating empty
        # routines when they should have rules in them, and this is causing
        # some feature files to fail to parse. :-/

        # if self.currentRoutine and not self.currentRoutine.rules:
        #     del(self.ff.routines[self.ff.routines.index(self.currentRoutine)])
        #     if self.currentFeature in self.ff.features:
        #         del(self.ff.features[self.currentFeature][-1])
        pass

    def add_feature_reference(self, location, featurename):
        # XXX
        pass

    def add_glyphClassDef(self, location, base, ligature, mark, component):
        for g in base:
            self.ff.glyphclasses[g] = "base"
        for g in ligature:
            self.ff.glyphclasses[g] = "ligature"
        for g in mark:
            self.ff.glyphclasses[g] = "mark"
        for g in component:
            self.ff.glyphclasses[g] = "component"



from .IndicShaperData import IndicPosition, make_syllable_machine, syllabic_category_map
from .SyllabicShaper import SyllabicShaper
from collections import OrderedDict


myanmar_category_reassignments = {
    0x002D: "GB",
    0x00A0: "GB",
    0x00D7: "GB",
    0x1004: "Ra",
    0x101B: "Ra",
    0x1032: "A",
    0x1036: "A",
    0x1038: "SM",
    0x1039: "H",
    0x103A: "As",
    0x103B: "MY",
    0x103C: "MR",
    0x103D: "MW",
    0x103E: "MH",
    0x1040: "D",  # XXX The spec says D0, but Uniscribe doesn't seem to do.
    0x104A: "P",
    0x104B: "P",
    0x104E: "C",
    0x105A: "Ra",
    0x105E: "MY",
    0x105F: "MY",
    0x1060: "MH",
    0x1082: "MW",
    0x109A: "SM",
    0x109B: "SM",
    0x109C: "SM",
    0x2012: "GB",
    0x2013: "GB",
    0x2014: "GB",
    0x2015: "GB",
    0x2022: "GB",
    0x25CC: "GB",
    0x25FB: "GB",
    0x25FC: "GB",
    0x25FD: "GB",
    0x25FE: "GB",
    0xAA74: "C",
    0xAA75: "C",
    0xAA76: "C",  # https://github.com/harfbuzz/harfbuzz/issues/218
}

for cp in range(0x1041, 0x1049 + 1):
    myanmar_category_reassignments[cp] = "D"
for cp in range(0x1090, 0x1099 + 1):
    myanmar_category_reassignments[cp] = "D"
for cp in range(0xF300, 0xFE0F + 1):
    myanmar_category_reassignments[cp] = "VS"
for cp in range(0x1087, 0x108F + 1):
    myanmar_category_reassignments[cp] = "SM"
for cp in [0x1063, 0x1064, 0x1069, 0x106A, 0x106B, 0x106C, 0x106D, 0xAA7B]:
    myanmar_category_reassignments[cp] = "PT"


states = OrderedDict(
    j="ZWJ|ZWNJ",  # Joiners
    k="(Ra As H)",  # Kinzi
    c="C|Ra",  # is_consonant
    medial_group="MY? As? MR? ((MW MH? | MH) As?)?",
    main_vowel_group="(VPre VS?)* VAbv* VBlw* A* (DB As?)?",
    post_vowel_group="VPst MH? As* VAbv* A* (DB As?)?",
    pwo_tone_group="PT A* DB? As?",
    complex_syllable_tail="As* medial_group main_vowel_group post_vowel_group* pwo_tone_group* V* j?",
    syllable_tail="(H (c|IV) VS?)* (H | complex_syllable_tail)",
    consonant_syllable="(k|CS)? (c|IV|D|GB) VS? syllable_tail",
    punctuation_cluster="P V",
    broken_cluster="k? VS? syllable_tail",
)


class MyanmarShaper(SyllabicShaper):
    basic_features = ["rphf", "pref", "blwf", "pstf"]
    other_features = ["pres", "abvs", "blws", "psts"]
    repha = None
    syllable_machine = make_syllable_machine(
        states,
        additional_categories=[
            "VS",
            "MW",
            "P",
            "As",
            "PT",
            "MY",
            "MH",
            "D",
            "GB",
            "MR",
            "VPre",
            "VAbv",
            "VBlw",
            "VPst",
        ],
    )
    syllable_types = [
        "consonant_syllable",
        "punctuation_cluster",
        "broken_cluster",
        "other",
    ]

    def reassign_category(self, item):
        cp = item.codepoint
        if cp in myanmar_category_reassignments:
            item.syllabic_category = myanmar_category_reassignments[cp]

        if item.syllabic_category == "M":
            if item.syllabic_position == IndicPosition.PRE_C:
                item.syllabic_category = "VPre"
                item.syllabic_position = IndicPosition.PRE_M
            elif item.syllabic_position == IndicPosition.ABOVE_C:
                item.syllabic_category = "VAbv"
            elif item.syllabic_position == IndicPosition.BELOW_C:
                item.syllabic_category = "VBlw"
            elif item.syllabic_position == IndicPosition.POST_C:
                item.syllabic_category = "VPst"

    def initial_reordering_consonant_syllable(self, start, end):
        def cat(i):
            return self.buffer.items[i].syllabic_category

        def get_pos(i):
            return self.buffer.items[i].syllabic_position

        def set_pos(i, pos):
            self.buffer.items[i].syllabic_position = pos

        def is_consonant(n):  # XXX Something else is Placeholder too
            isc = cat(n)
            is_medial = isc == "CM"
            return (
                isc in ["C", "CS", "Ra", "V", "PLACEHOLDER", "DOTTEDCIRCLE"]
                or is_medial
            )

        base = end
        has_reph = False

        limit = start
        if (
            start + 3 <= end
            and cat(start) == "Ra"
            and cat(start + 1) == "As"
            and cat(start + 2) == "H"
        ):
            limit += 3
            base = start
            has_reph = True
        if not has_reph:
            base = limit
        for i in range(limit, end):
            if is_consonant(i):
                base = i
                break
        i = start
        while i < start + (3 if has_reph else 0):
            set_pos(i, IndicPosition.AFTER_MAIN)
            i += 1
        while i < base:
            set_pos(i, IndicPosition.PRE_C)
            i += 1
        if i < end:
            set_pos(i, IndicPosition.BASE_C)
            i += 1
        pos = IndicPosition.AFTER_MAIN
        while i < end:
            if cat(i) == "MR":
                set_pos(i, IndicPosition.PRE_C)
            elif get_pos(i) < IndicPosition.BASE_C:  # Left Matra
                pass
            elif cat(i) == "VS":
                set_pos(i, get_pos(i - 1))
            elif pos == IndicPosition.AFTER_MAIN and cat(i) == "VBlw":
                pos = IndicPosition.BELOW_C
                set_pos(i, pos)
            elif pos == IndicPosition.BELOW_C and cat(i) == "A":
                set_pos(i, IndicPosition.BEFORE_SUB)
            elif pos == IndicPosition.BELOW_C and cat(i) == "VBlw":
                set_pos(i, pos)
            elif pos == IndicPosition.BELOW_C and cat(i) != "A":
                pos = IndicPosition.AFTER_SUB
                set_pos(i, pos)
            else:
                set_pos(i, pos)
            i = i + 1
            continue

        self.buffer.items[start:end] = sorted(
            self.buffer.items[start:end], key=lambda x: x.syllabic_position
        )

    initial_reordering_syllable = {
        "broken_cluster": initial_reordering_consonant_syllable,
        "consonant_syllable": initial_reordering_consonant_syllable,
    }



"""GPOSUnparser: Convert binary GPOS lookups to fontFeatures objects."""
from .GTableUnparser import GTableUnparser
import fontFeatures


class GPOSUnparser(GTableUnparser):
    """Unparse a GPOS table into a fontFeatures object. See :py:class:`fontFeatures.ttLib.GTableUnparser`."""

    _table = "GPOS"
    lookupTypes = {
        1: "SinglePositioning",
        2: "PairPositioning",
        3: "CursiveAttachment",
        4: "MarkToBase",
        5: "MarkToLigature",
        6: "MarkToMark",
        7: "Contextual",
        8: "ChainedContextual",
        9: "Extension",
    }

    _attrs = {
        "lookup": "PosLookupRecord",
        "format1_ruleset": "PosRuleSet",
        "format1_rule": "PosRule",
        "format2_classset": "PosClassSet",
        "format2_rule": "PosClassRule",
        "chain_format1_ruleset": "ChainPosRuleSet",
        "chain_format1_rule": "ChainPosRule",
        "chain_format2_classset": "ChainPosClassSet",
        "chain_format2_rule": "ChainPosClassRule",
    }

    def makeValueRecord(self, valueRecord):
        """Helper routine to create ValueRecord instances.

        Args:
            valueRecord: An ``otTables.ValueRecord`` object.

        Returns a ``fontFeatures.ValueRecord`` object."""
        valueFormatFlags = (
            ("XPlacement", 0x0001),  # Includes horizontal adjustment for placement
            ("YPlacement", 0x0002),  # Includes vertical adjustment for placement
            ("XAdvance", 0x0004),  # Includes horizontal adjustment for advance
            ("YAdvance", 0x0008),  # Includes vertical adjustment for advance
            # Currently we don't have a way to express or represent value records
            # with Device tables, whether old-style (ppem adjustment) or new-style
            # (VariationIndex). Even if we could represent them, then what? See
            # issue #31.
            # ("XPlaDevice", 0x0010),  # Includes horizontal Device table for placement
            # ("YPlaDevice", 0x0020),  # Includes vertical Device table for placement
            # ("XAdvDevice", 0x0040),  # Includes horizontal Device table for advance
            # ("YAdvDevice", 0x0080)  # Includes vertical Device table for advance
            # , 'Reserved': 0xF000 For future use (set to zero)
        )

        # defaults to 0
        values = [getattr(valueRecord, name, 0) or None for name, _ in valueFormatFlags]
        return fontFeatures.ValueRecord(*values)

    def isChaining(self, lookupType):
        """Returns true if the given lookup type is a chaining lookup."""
        return lookupType >= 7

    def unparseSinglePositioning(self, lookup):
        """Turn a GPOS1 (single positioning) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("SinglePositioning" + self.gensym()))
        self._fix_flags(b, lookup)

        for subtable in lookup.SubTable:
            if subtable.Format == 1:
                spos = fontFeatures.Positioning(
                    [subtable.Coverage.glyphs],
                    [self.makeValueRecord(subtable.Value)],
                    address=self.currentLookup,
                    flags=lookup.LookupFlag,
                )
                b.addRule(spos)
            else:
                # Optimize it later
                for g, v in zip(subtable.Coverage.glyphs, subtable.Value):
                    spos = fontFeatures.Positioning(
                        [[g]],
                        [self.makeValueRecord(v)],
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                    b.addRule(spos)
        return b, []

    def unparsePairPositioning(self, lookup):
        """Turn a GPOS2 (pair adjustment) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("PairPositioning" + self.gensym()))
        self._fix_flags(b, lookup)
        for subtable in lookup.SubTable:
            if subtable.Format == 1:
                for g, pair in zip(subtable.Coverage.glyphs, subtable.PairSet):
                    for vr in pair.PairValueRecord:
                        spos = fontFeatures.Positioning(
                            [[g], [vr.SecondGlyph]],
                            [
                                self.makeValueRecord(vr.Value1),
                                self.makeValueRecord(vr.Value2),
                            ],
                            address=self.currentLookup,
                            flags=lookup.LookupFlag,
                        )
                        b.addRule(spos)
            else:
                class1 = self._invertClassDef(subtable.ClassDef1.classDefs, self.font)
                class2 = self._invertClassDef(subtable.ClassDef2.classDefs, self.font)
                for ix1, c1 in enumerate(subtable.Class1Record):
                    if ix1 not in class1:
                        continue  # XXX
                    for ix2, c2 in enumerate(c1.Class2Record):
                        if ix2 not in class2:
                            continue  # XXX
                        vr1 = self.makeValueRecord(c2.Value1)
                        vr2 = self.makeValueRecord(c2.Value2)
                        if not vr1 and not vr2:
                            continue
                        firstClass = list(
                            set(class1[ix1]) & set(subtable.Coverage.glyphs)
                        )
                        spos = fontFeatures.Positioning(
                            [firstClass, class2[ix2]],
                            [vr1, vr2],
                            address=self.currentLookup,
                            flags=lookup.LookupFlag,
                        )
                        b.addRule(spos)
        return b, []

    def unparseCursiveAttachment(self, lookup):
        """Turn a GPOS3 (cursive attachment) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("CursiveAttachment" + self.gensym()))
        self._fix_flags(b, lookup)
        entries = {}
        exits = {}
        for s in lookup.SubTable:
            assert s.Format == 1
            for glyph, record in zip(s.Coverage.glyphs, s.EntryExitRecord):
                if record.EntryAnchor:
                    entries[glyph] = (
                        record.EntryAnchor.XCoordinate,
                        record.EntryAnchor.YCoordinate,
                    )
                if record.ExitAnchor:
                    exits[glyph] = (
                        record.ExitAnchor.XCoordinate,
                        record.ExitAnchor.YCoordinate,
                    )
        b.addRule(
            fontFeatures.Attachment(
                "cursive_entry",
                "cursive_exit",
                entries,
                exits,
                flags=lookup.LookupFlag,
                address=self.currentLookup,
            )
        )
        return b, []

    def unparseMarkToBase(self, lookup):
        """Turn a GPOS4 (mark to base) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("MarkToBase" + self.gensym()))
        self._fix_flags(b, lookup)
        for subtable in lookup.SubTable:  # fontTools.ttLib.tables.otTables.MarkBasePos
            assert subtable.Format == 1
            for classId in range(0, subtable.ClassCount):
                anchorClassPrefix = "Anchor" + self.gensym()
                marks = self._formatMarkArray(
                    subtable.MarkArray, subtable.MarkCoverage, classId
                )
                bases = self._formatBaseArray(
                    subtable.BaseArray, subtable.BaseCoverage, classId
                )
                b.addRule(
                    fontFeatures.Attachment(
                        anchorClassPrefix,
                        "_"+anchorClassPrefix,
                        bases,
                        marks,
                        font=self.font,
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                    )
                )
        return b, []

    def _formatMarkArray(self, markArray, markCoverage, classId):
        id2Name = markCoverage.glyphs
        marks = {}
        for i, markRecord in enumerate(markArray.MarkRecord):
            if markRecord.Class == classId:
                marks[id2Name[i]] = (
                    markRecord.MarkAnchor.XCoordinate,
                    markRecord.MarkAnchor.YCoordinate,
                    # markRecord.Class
                )
        return marks

    def _formatMark2Array(self, markArray, markCoverage, anchorClassPrefix):
        id2Name = markCoverage.glyphs
        marks = {}
        for i, markRecord in enumerate(markArray.Mark2Record):
            anchor = markRecord.Mark2Anchor[anchorClassPrefix]
            if not anchor:
                continue
            marks[id2Name[i]] = (
                anchor.XCoordinate,
                anchor.YCoordinate,
            )
        return marks

    def _formatBaseArray(self, baseArray, baseCoverage, wantedClassId):
        id2Name = baseCoverage.glyphs
        bases = {}
        for i, baseRecord in enumerate(baseArray.BaseRecord):
            for classId, anchor in enumerate(baseRecord.BaseAnchor):
                if classId != wantedClassId:
                    continue
                if not anchor:
                    continue
                bases[id2Name[i]] = (anchor.XCoordinate, anchor.YCoordinate)  # ClassId?
        return bases

    def unparseMarkToLigature(self, lookup):
        """Turn a GPOS5 (mark to ligature) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("MarkToLigature" + self.gensym()))
        self._fix_flags(b, lookup)
        self.unparsable(b, "Mark to lig pos", lookup)
        return b, []

    def unparseMarkToMark(self, lookup):
        """Turn a GPOS6 (mark to mark) subtable into a fontFeatures Routine."""
        b = fontFeatures.Routine(name=self.getname("MarkToMark" + self.gensym()))
        self._fix_flags(b, lookup)
        for subtable in lookup.SubTable:  # fontTools.ttLib.tables.otTables.MarkBasePos
            assert subtable.Format == 1
            for classId in range(0, subtable.ClassCount):
                anchorClassPrefix = "Anchor" + self.gensym()
                marks = self._formatMarkArray(
                    subtable.Mark1Array, subtable.Mark1Coverage, classId
                )
                bases = self._formatMark2Array(
                    subtable.Mark2Array, subtable.Mark2Coverage, classId
                )
                b.addRule(
                    fontFeatures.Attachment(
                        anchorClassPrefix,
                        anchorClassPrefix + "_",
                        bases,
                        marks,
                        font=self.font,
                        address=self.currentLookup,
                        flags=lookup.LookupFlag,
                        force_markmark=True
                    )
                )
        return b, []



from fontFeatures import Substitution, FontFeatures, Routine, Chaining, RoutineReference
from fontFeatures.feaLib import FeaParser
from fontTools.ttLib import TTFont
import pytest
from fontTools.misc.testTools import getXML


font = TTFont("fonts/Roboto-Regular.ttf")


def test_script_language_split():
    f = FontFeatures()

    s1 = Substitution([["question"]],  [["questiongreek"]], languages=[ ("grek","*") ])
    s2 = Substitution([["A"]],  [["alpha"]], languages=[ ("grek","ELL ") ])
    s3 = Substitution([["question"]],  [["questiondown"]], languages=[ ("latn", "ESP ") ])
    s4 = Substitution([["question"]],  [["B"]], languages=[ ("latn", "POL ") ])
    s5 = Substitution([["X"]],  [["Y"]])
    r = Routine(rules=[s1,s2,s3,s4,s5])

    f.addFeature("locl", [r])
    f.buildBinaryFeatures(font)
    gsub = "\n".join(getXML(font["GSUB"].toXML))
    assert gsub == """<Version value="0x00010000"/>
<ScriptList>
  <!-- ScriptCount=3 -->
  <ScriptRecord index="0">
    <ScriptTag value="DFLT"/>
    <Script>
      <DefaultLangSys>
        <ReqFeatureIndex value="65535"/>
        <!-- FeatureCount=1 -->
        <FeatureIndex index="0" value="3"/>
      </DefaultLangSys>
      <!-- LangSysCount=0 -->
    </Script>
  </ScriptRecord>
  <ScriptRecord index="1">
    <ScriptTag value="grek"/>
    <Script>
      <DefaultLangSys>
        <ReqFeatureIndex value="65535"/>
        <!-- FeatureCount=1 -->
        <FeatureIndex index="0" value="4"/>
      </DefaultLangSys>
      <!-- LangSysCount=1 -->
      <LangSysRecord index="0">
        <LangSysTag value="ELL "/>
        <LangSys>
          <ReqFeatureIndex value="65535"/>
          <!-- FeatureCount=1 -->
          <FeatureIndex index="0" value="0"/>
        </LangSys>
      </LangSysRecord>
    </Script>
  </ScriptRecord>
  <ScriptRecord index="2">
    <ScriptTag value="latn"/>
    <Script>
      <DefaultLangSys>
        <ReqFeatureIndex value="65535"/>
        <!-- FeatureCount=1 -->
        <FeatureIndex index="0" value="3"/>
      </DefaultLangSys>
      <!-- LangSysCount=2 -->
      <LangSysRecord index="0">
        <LangSysTag value="ESP "/>
        <LangSys>
          <ReqFeatureIndex value="65535"/>
          <!-- FeatureCount=1 -->
          <FeatureIndex index="0" value="1"/>
        </LangSys>
      </LangSysRecord>
      <LangSysRecord index="1">
        <LangSysTag value="POL "/>
        <LangSys>
          <ReqFeatureIndex value="65535"/>
          <!-- FeatureCount=1 -->
          <FeatureIndex index="0" value="2"/>
        </LangSys>
      </LangSysRecord>
    </Script>
  </ScriptRecord>
</ScriptList>
<FeatureList>
  <!-- FeatureCount=5 -->
  <FeatureRecord index="0">
    <FeatureTag value="locl"/>
    <Feature>
      <!-- LookupCount=3 -->
      <LookupListIndex index="0" value="0"/>
      <LookupListIndex index="1" value="1"/>
      <LookupListIndex index="2" value="4"/>
    </Feature>
  </FeatureRecord>
  <FeatureRecord index="1">
    <FeatureTag value="locl"/>
    <Feature>
      <!-- LookupCount=2 -->
      <LookupListIndex index="0" value="2"/>
      <LookupListIndex index="1" value="4"/>
    </Feature>
  </FeatureRecord>
  <FeatureRecord index="2">
    <FeatureTag value="locl"/>
    <Feature>
      <!-- LookupCount=2 -->
      <LookupListIndex index="0" value="3"/>
      <LookupListIndex index="1" value="4"/>
    </Feature>
  </FeatureRecord>
  <FeatureRecord index="3">
    <FeatureTag value="locl"/>
    <Feature>
      <!-- LookupCount=1 -->
      <LookupListIndex index="0" value="4"/>
    </Feature>
  </FeatureRecord>
  <FeatureRecord index="4">
    <FeatureTag value="locl"/>
    <Feature>
      <!-- LookupCount=2 -->
      <LookupListIndex index="0" value="0"/>
      <LookupListIndex index="1" value="4"/>
    </Feature>
  </FeatureRecord>
</FeatureList>
<LookupList>
  <!-- LookupCount=5 -->
  <Lookup index="0">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="question" out="questiongreek"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="1">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="A" out="alpha"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="2">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="question" out="questiondown"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="3">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="question" out="B"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="4">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="X" out="Y"/>
    </SingleSubst>
  </Lookup>
</LookupList>"""


def test_different_sub_types():
    parser = FeaParser("""
lookup single { sub a by b; } single;
lookup single_flags { lookupflag IgnoreMarks; sub a by b; } single_flags;
lookup multiple { sub a by b c; } multiple;
lookup ligature { sub a b by c; } ligature;
lookup chain1 { sub a b' by c; } chain1;
lookup rsub { rsub a b' c by d; } rsub;
lookup chain2 { sub a' lookup single_flags b; } chain2;
""")
    parser.parse()
    parser.ff.buildBinaryFeatures(font)
    gsub = "\n".join(getXML(font["GSUB"].toXML))
    assert gsub == """<Version value="0x00010000"/>
<ScriptList>
  <!-- ScriptCount=0 -->
</ScriptList>
<FeatureList>
  <!-- FeatureCount=0 -->
</FeatureList>
<LookupList>
  <!-- LookupCount=8 -->
  <Lookup index="0">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="a" out="b"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="1">
    <LookupType value="1"/>
    <LookupFlag value="8"/><!-- ignoreMarks -->
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="a" out="b"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="2">
    <LookupType value="2"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <MultipleSubst index="0">
      <Substitution in="a" out="b,c"/>
    </MultipleSubst>
  </Lookup>
  <Lookup index="3">
    <LookupType value="4"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <LigatureSubst index="0">
      <LigatureSet glyph="a">
        <Ligature components="b" glyph="c"/>
      </LigatureSet>
    </LigatureSubst>
  </Lookup>
  <Lookup index="4">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SingleSubst index="0">
      <Substitution in="b" out="c"/>
    </SingleSubst>
  </Lookup>
  <Lookup index="5">
    <LookupType value="6"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <ChainContextSubst index="0" Format="3">
      <!-- BacktrackGlyphCount=1 -->
      <BacktrackCoverage index="0">
        <Glyph value="a"/>
      </BacktrackCoverage>
      <!-- InputGlyphCount=1 -->
      <InputCoverage index="0">
        <Glyph value="b"/>
      </InputCoverage>
      <!-- LookAheadGlyphCount=0 -->
      <!-- SubstCount=1 -->
      <SubstLookupRecord index="0">
        <SequenceIndex value="0"/>
        <LookupListIndex value="4"/>
      </SubstLookupRecord>
    </ChainContextSubst>
  </Lookup>
  <Lookup index="6">
    <LookupType value="8"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <ReverseChainSingleSubst index="0" Format="1">
      <Coverage>
        <Glyph value="b"/>
      </Coverage>
      <!-- BacktrackGlyphCount=1 -->
      <BacktrackCoverage index="0">
        <Glyph value="a"/>
      </BacktrackCoverage>
      <!-- LookAheadGlyphCount=1 -->
      <LookAheadCoverage index="0">
        <Glyph value="c"/>
      </LookAheadCoverage>
      <!-- GlyphCount=1 -->
      <Substitute index="0" value="d"/>
    </ReverseChainSingleSubst>
  </Lookup>
  <Lookup index="7">
    <LookupType value="6"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <ChainContextSubst index="0" Format="3">
      <!-- BacktrackGlyphCount=0 -->
      <!-- InputGlyphCount=1 -->
      <InputCoverage index="0">
        <Glyph value="a"/>
      </InputCoverage>
      <!-- LookAheadGlyphCount=1 -->
      <LookAheadCoverage index="0">
        <Glyph value="b"/>
      </LookAheadCoverage>
      <!-- SubstCount=1 -->
      <SubstLookupRecord index="0">
        <SequenceIndex value="0"/>
        <LookupListIndex value="1"/>
      </SubstLookupRecord>
    </ChainContextSubst>
  </Lookup>
</LookupList>"""

def test_different_pos_types():
    parser = FeaParser("""
markClass [acute grave] <anchor 150 -10> @TOP_MARKS;
lookup single { pos a 50; } single;
lookup pair_and_class_pair { pos a 20 b 50; pos [a b] 15 c 10; } pair_and_class_pair;
lookup cursive { pos cursive c <anchor 500 200> <anchor 0 0>; } cursive;
lookup markbase { position base [a b] <anchor 250 450> mark @TOP_MARKS; } markbase;
lookup markmark { position mark [dieresis] <anchor 20 40> mark @TOP_MARKS; } markmark;
lookup chain1 { pos a' 50 b; } chain1;
lookup chain2 { pos a' lookup single b; } chain2;
""")
    parser.parse()
    parser.ff.glyphclasses["dieresis"] = "mark"
    parser.ff.buildBinaryFeatures(font)
    gpos = "\n".join(getXML(font["GPOS"].toXML))
    assert gpos == """<Version value="0x00010000"/>
<ScriptList>
  <!-- ScriptCount=0 -->
</ScriptList>
<FeatureList>
  <!-- FeatureCount=0 -->
</FeatureList>
<LookupList>
  <!-- LookupCount=8 -->
  <Lookup index="0">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SinglePos index="0" Format="1">
      <Coverage>
        <Glyph value="a"/>
      </Coverage>
      <ValueFormat value="4"/>
      <Value XAdvance="50"/>
    </SinglePos>
  </Lookup>
  <Lookup index="1">
    <LookupType value="2"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=2 -->
    <PairPos index="0" Format="1">
      <Coverage>
        <Glyph value="a"/>
      </Coverage>
      <ValueFormat1 value="4"/>
      <ValueFormat2 value="4"/>
      <!-- PairSetCount=1 -->
      <PairSet index="0">
        <!-- PairValueCount=1 -->
        <PairValueRecord index="0">
          <SecondGlyph value="b"/>
          <Value1 XAdvance="20"/>
          <Value2 XAdvance="50"/>
        </PairValueRecord>
      </PairSet>
    </PairPos>
    <PairPos index="1" Format="2">
      <Coverage>
        <Glyph value="a"/>
        <Glyph value="b"/>
      </Coverage>
      <ValueFormat1 value="4"/>
      <ValueFormat2 value="4"/>
      <ClassDef1>
      </ClassDef1>
      <ClassDef2>
        <ClassDef glyph="c" class="1"/>
      </ClassDef2>
      <!-- Class1Count=1 -->
      <!-- Class2Count=2 -->
      <Class1Record index="0">
        <Class2Record index="0">
          <Value1 XAdvance="0"/>
          <Value2 XAdvance="0"/>
        </Class2Record>
        <Class2Record index="1">
          <Value1 XAdvance="15"/>
          <Value2 XAdvance="10"/>
        </Class2Record>
      </Class1Record>
    </PairPos>
  </Lookup>
  <Lookup index="2">
    <LookupType value="3"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <CursivePos index="0" Format="1">
      <Coverage>
        <Glyph value="c"/>
      </Coverage>
      <!-- EntryExitCount=1 -->
      <EntryExitRecord index="0">
        <EntryAnchor Format="1">
          <XCoordinate value="500"/>
          <YCoordinate value="200"/>
        </EntryAnchor>
        <ExitAnchor Format="1">
          <XCoordinate value="0"/>
          <YCoordinate value="0"/>
        </ExitAnchor>
      </EntryExitRecord>
    </CursivePos>
  </Lookup>
  <Lookup index="3">
    <LookupType value="4"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <MarkBasePos index="0" Format="1">
      <MarkCoverage>
        <Glyph value="grave"/>
        <Glyph value="acute"/>
      </MarkCoverage>
      <BaseCoverage>
        <Glyph value="a"/>
        <Glyph value="b"/>
      </BaseCoverage>
      <!-- ClassCount=1 -->
      <MarkArray>
        <!-- MarkCount=2 -->
        <MarkRecord index="0">
          <Class value="0"/>
          <MarkAnchor Format="1">
            <XCoordinate value="150"/>
            <YCoordinate value="-10"/>
          </MarkAnchor>
        </MarkRecord>
        <MarkRecord index="1">
          <Class value="0"/>
          <MarkAnchor Format="1">
            <XCoordinate value="150"/>
            <YCoordinate value="-10"/>
          </MarkAnchor>
        </MarkRecord>
      </MarkArray>
      <BaseArray>
        <!-- BaseCount=2 -->
        <BaseRecord index="0">
          <BaseAnchor index="0" Format="1">
            <XCoordinate value="250"/>
            <YCoordinate value="450"/>
          </BaseAnchor>
        </BaseRecord>
        <BaseRecord index="1">
          <BaseAnchor index="0" Format="1">
            <XCoordinate value="250"/>
            <YCoordinate value="450"/>
          </BaseAnchor>
        </BaseRecord>
      </BaseArray>
    </MarkBasePos>
  </Lookup>
  <Lookup index="4">
    <LookupType value="6"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <MarkMarkPos index="0" Format="1">
      <Mark1Coverage>
        <Glyph value="grave"/>
        <Glyph value="acute"/>
      </Mark1Coverage>
      <Mark2Coverage>
        <Glyph value="dieresis"/>
      </Mark2Coverage>
      <!-- ClassCount=1 -->
      <Mark1Array>
        <!-- MarkCount=2 -->
        <MarkRecord index="0">
          <Class value="0"/>
          <MarkAnchor Format="1">
            <XCoordinate value="150"/>
            <YCoordinate value="-10"/>
          </MarkAnchor>
        </MarkRecord>
        <MarkRecord index="1">
          <Class value="0"/>
          <MarkAnchor Format="1">
            <XCoordinate value="150"/>
            <YCoordinate value="-10"/>
          </MarkAnchor>
        </MarkRecord>
      </Mark1Array>
      <Mark2Array>
        <!-- Mark2Count=1 -->
        <Mark2Record index="0">
          <Mark2Anchor index="0" Format="1">
            <XCoordinate value="20"/>
            <YCoordinate value="40"/>
          </Mark2Anchor>
        </Mark2Record>
      </Mark2Array>
    </MarkMarkPos>
  </Lookup>
  <Lookup index="5">
    <LookupType value="1"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <SinglePos index="0" Format="1">
      <Coverage>
        <Glyph value="a"/>
      </Coverage>
      <ValueFormat value="4"/>
      <Value XAdvance="50"/>
    </SinglePos>
  </Lookup>
  <Lookup index="6">
    <LookupType value="8"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <ChainContextPos index="0" Format="3">
      <!-- BacktrackGlyphCount=0 -->
      <!-- InputGlyphCount=1 -->
      <InputCoverage index="0">
        <Glyph value="a"/>
      </InputCoverage>
      <!-- LookAheadGlyphCount=1 -->
      <LookAheadCoverage index="0">
        <Glyph value="b"/>
      </LookAheadCoverage>
      <!-- PosCount=1 -->
      <PosLookupRecord index="0">
        <SequenceIndex value="0"/>
        <LookupListIndex value="5"/>
      </PosLookupRecord>
    </ChainContextPos>
  </Lookup>
  <Lookup index="7">
    <LookupType value="8"/>
    <LookupFlag value="0"/>
    <!-- SubTableCount=1 -->
    <ChainContextPos index="0" Format="3">
      <!-- BacktrackGlyphCount=0 -->
      <!-- InputGlyphCount=1 -->
      <InputCoverage index="0">
        <Glyph value="a"/>
      </InputCoverage>
      <!-- LookAheadGlyphCount=1 -->
      <LookAheadCoverage index="0">
        <Glyph value="b"/>
      </LookAheadCoverage>
      <!-- PosCount=1 -->
      <PosLookupRecord index="0">
        <SequenceIndex value="0"/>
        <LookupListIndex value="0"/>
      </PosLookupRecord>
    </ChainContextPos>
  </Lookup>
</LookupList>"""



from fontFeatures import Substitution, FontFeatures, Routine

import pytest


def test_language_ordering():
    f = FontFeatures()
    s1 = Substitution([["a"]], ["b"], languages=[("arab", "URD ")])
    s2 = Substitution([["a"]], ["c"], languages=[("arab", "FAR ")])
    s3 = Substitution([["x"], ["y"]], ["z"], languages=[("arab", "URD ")])

    f.addFeature("locl", [Routine(rules=[ s1, s2, s3 ] )])

    # When going to fea, we put everything in its own feature block to
    # avoid stupid problems with language systems
    expected = """languagesystem arab URD;
languagesystem arab FAR;

lookup Routine_1 {
    lookupflag 0;
    ;
    sub a by b;
} Routine_1;

lookup Routine_2 {
    lookupflag 0;
    ;
    sub a by c;
} Routine_2;

lookup Routine_3 {
    lookupflag 0;
    ;
    sub x y by z;
} Routine_3;

feature locl {
    script arab;
    language URD;
            lookup Routine_1;

} locl;

feature locl {
    script arab;
    language FAR;
            lookup Routine_2;

} locl;

feature locl {
    script arab;
    language URD;
            lookup Routine_3;

} locl;
"""
    assert f.asFea(do_gdef=False) == expected

    # But the same is true of multiple lookups in the same feature
    f = FontFeatures()
    r1 = Routine(rules=[ s1 ])
    r2 = Routine(rules=[ s2 ])
    r3 = Routine(rules=[ s3 ])

    f.addFeature("locl", [r1,r2,r3])
    assert f.asFea(do_gdef=False) == expected


def test_multiple_languages():
    f = FontFeatures()
    s1 = Substitution([["a"]], ["b"], languages=[("arab", "URD "),("arab", "FAR ")])
    expected = """languagesystem arab URD;
languagesystem arab FAR;

lookup Routine_1 {
    lookupflag 0;
    ;
    sub a by b;
} Routine_1;

lookup Routine_2 {
    lookupflag 0;
    ;
    sub a by b;
} Routine_2;

feature locl {
    script arab;
    language URD;
            lookup Routine_1;

} locl;

feature locl {
    script arab;
    language FAR;
            lookup Routine_2;

} locl;
"""
    f = FontFeatures()
    r1 = Routine(rules=[ s1 ])
    f.addFeature("locl", [r1])
    assert f.asFea(do_gdef=False) == expected

def test_multiple_languages_routine():
    f = FontFeatures()
    s1 = Substitution([["a"]], ["b"])
    expected = """languagesystem arab URD;
languagesystem arab FAR;

lookup Routine_1 {
    lookupflag 0;
    ;
    sub a by b;
} Routine_1;

feature locl {
    script arab;
    language URD;
            lookup Routine_1;

} locl;

feature locl {
    script arab;
    language FAR;
            lookup Routine_1;

} locl;
"""
    f = FontFeatures()
    r1 = Routine(rules=[ s1 ], languages=[("arab", "URD "),("arab", "FAR ")])
    f.addFeature("locl", [r1])
    assert f.asFea(do_gdef=False) == expected




"""Routines for converting Chaining rules to and from XML."""

from lxml import etree


def _toXML(self, root):
    self._makeglyphslots(root, "input", self.input)
    return root


@classmethod
def fromXML(klass, el):
  """Creates a rule from a lxml Element object."""
  from fontFeatures import Routine, RoutineReference
  rule = klass(
      klass._slotArray(klass, el.find("input")),
      precontext = klass._slotArray(klass, el.find("precontext")),
      postcontext = klass._slotArray(klass, el.find("postcontext")),
      address = el.get("address"),
      languages = el.get("languages"),
      flags = int(el.get("flags") or 0)
  )
  lookupsxml = el.find("lookups")
  rule.lookups = []
  for slot in lookupsxml:
    routines = []
    for lu in slot:
        if lu.tag == "routinereference":
            routines.append(RoutineReference.fromXML(lu))
        elif lu.tag == "routine":
            routines.append(Routine.fromXML(lu))
    rule.lookups.append(routines)
  return rule



def shaper_inputs(self):
    return self.input

def __find_masked_ix(buf, ix):
    # This probably could be done better...
    for i in range(len(buf.items)):
        if buf.mask[i] == ix: return i
    return -1

def _do_apply(self, buf, ix, namedclasses={}):
    from fontFeatures import RoutineReference
    # Save buffer mask
    flags = buf.flags
    markFilteringSet = buf.markFilteringSet
    markAttachmentSet = buf.markAttachmentSet

    if ix + len(self.lookups) -1 > len(buf.mask):
        return

    old_unmasked_indexes = [ buf.mask[ix+i] for i in range(len(self.lookups)) ]
    for i,lookups in enumerate(self.lookups):
        if not lookups:
            continue
        for routine in lookups:
            assert isinstance(routine, RoutineReference)
            routine = routine.routine
            # Adjust mask and recompute index?
            unmasked_ix = old_unmasked_indexes[i]
            for rule in routine.rules:
                buf.set_mask(rule.flags, routine.markFilteringSet, routine.markAttachmentSet)
                newix = __find_masked_ix(buf, unmasked_ix)
                if rule.would_apply_at_position(buf,newix, namedclasses) and rule._do_apply(buf, newix):
                    break

    buf.set_mask(flags, markFilteringSet, markAttachmentSet)
    return len(self.input) - 1




# Code for converting a Chaining object into feaLib statements
import fontTools.feaLib.ast as feaast
import fontFeatures


# Can we chain multiple lookups?
from fontTools.feaLib.parser import Parser
import io


def glyphref(g):
    if len(g) == 1:
        return feaast.GlyphName(g[0])
    return feaast.GlyphClass([feaast.GlyphName(x) for x in g])



def gensym(ff):
    if not "index" in ff.scratch:
        ff.scratch["index"] = 0
    ff.scratch["index"] = ff.scratch["index"] + 1
    return str(ff.scratch["index"])


def replaceLongWithClasses(i, ff):
    for ix, gc in enumerate(i):
        if len(gc) > 5:
            classname = ff.getNamedClassFor(sorted(gc), "class" + gensym(ff))
            i[ix] = ["@" + classname]


def feaPreamble(self, ff):
    if not "glyphclasses" in ff.scratch:
        ff.scratch["glyphclasses"] = {tuple(sorted(ff.namedClasses[g])): g for g in ff.namedClasses.keys()}
    replaceLongWithClasses(self.input, ff)
    replaceLongWithClasses(self.precontext, ff)
    replaceLongWithClasses(self.postcontext, ff)
    from fontFeatures import RoutineReference

    # Ensure all linked routines have names
    for lul in self.lookups:
        for r in (lul or []):
            assert isinstance(r, RoutineReference)
            assert r.routine
            if not r.routine.name:
                r.routine.name = "ChainedRoutine"+gensym(ff)
            if not r.name:
                r.name = r.routine.name

    return []


def _complex(self):
    if self.stage == "sub":
        routine = feaast.ChainContextSubstStatement
    else:
        routine = feaast.ChainContextPosStatement

    return routine(
        [glyphref(x) for x in self.precontext],
        [glyphref(x) for x in self.input],
        [glyphref(x) for x in self.postcontext],
        self.lookups,
    )


def asFeaAST(self):
    if len(self.lookups) > 0 and any([x is not None for x in self.lookups]):
        # Fill in the blanks
        if self.stage == "sub":
            routine = feaast.ChainContextSubstStatement
        else:
            routine = feaast.ChainContextPosStatement
        # Check for >1 lookups per position
        if any([x and len(x) > 1 for x in self.lookups]):
            return _complex(self)
        lookups = self.lookups
        return routine(
            [glyphref(x) for x in self.precontext],
            [glyphref(x) for x in self.input],
            [glyphref(x) for x in self.postcontext],
            lookups,
        )
    else:
        return feaast.IgnoreSubstStatement(
            chainContexts=[
                [
                    [glyphref(x) for x in self.precontext],
                    [glyphref(x) for x in self.input],
                    [glyphref(x) for x in self.postcontext],
                ]
            ]
        )




"""ttLib.Positioning: Converting Positioning rules to TrueType."""

def lookup_type(self):
    """Mixin to determine the GPOS lookup type of a fontFeatures.Positioning object

    Returns: integer GPOS lookup type."""
    if not self.has_context:
        if len(self.glyphs) == 1:
            return 1
        if len(self.glyphs) == 2:
            return 2
    return 8




# The following functions are generated by running:
# utils/gen-vowel-constraints.py ms-use/IndicShapingInvalidCluster.txt

from fontFeatures.shaperLib.Buffer import BufferItem

DOTTED_CIRCLE = 0x25CC

def _insert_dotted_circle(buf, index):
    dotted_circle = BufferItem.new_unicode(DOTTED_CIRCLE)
    buf.items.insert(index, dotted_circle)


def preprocess_text_vowel_constraints(buffer):
    if buffer.script == "Devanagari":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0905:
                    if buffer.items[i+1].codepoint in [0x093A, 0x093B, 0x093E, 0x0945, 0x0946, 0x0949, 0x094A, 0x094B, 0x094C, 0x094F, 0x0956, 0x0957]:
                        matched = True
            if buffer.items[i].codepoint == 0x0906:
                    if buffer.items[i+1].codepoint in [0x093A, 0x0945, 0x0946, 0x0947, 0x0948]:
                        matched = True
            if buffer.items[i].codepoint == 0x0909:
                matched = 0x0941 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x090F:
                    if buffer.items[i+1].codepoint in [0x0945, 0x0946, 0x0947]:
                        matched = True
            if buffer.items[i].codepoint == 0x0930:
                if (0x094D == buffer.items[i+1].codepoint and
                        i + 2 < len(buffer.items) and
                        0x0907 == buffer.items[i+2].codepoint):
                    i = i + 1
                    matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Bengali":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0985:
                matched = 0x09BE == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x098B:
                matched = 0x09C3 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x098C:
                matched = 0x09E2 == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Gurmukhi":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0A05:
                    if buffer.items[i+1].codepoint in [0x0A3E, 0x0A48, 0x0A4C]:
                        matched = True
            if buffer.items[i].codepoint == 0x0A72:
                    if buffer.items[i+1].codepoint in [0x0A3F, 0x0A40, 0x0A47]:
                        matched = True
            if buffer.items[i].codepoint == 0x0A73:
                    if buffer.items[i+1].codepoint in [0x0A41, 0x0A42, 0x0A4B]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Gujarati":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0A85:
                    if buffer.items[i+1].codepoint in [0x0ABE, 0x0AC5, 0x0AC7, 0x0AC8, 0x0AC9, 0x0ACB, 0x0ACC]:
                        matched = True
            if buffer.items[i].codepoint == 0x0AC5:
                matched = 0x0ABE == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Oriya":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0B05:
                matched = 0x0B3E == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint in [0x0B0F, 0x0B13]:
                matched = 0x0B57 == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Tamil":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if (0x0B85 == buffer.items[i].codepoint and
                    0x0BC2 == buffer.items[i+1].codepoint):
                matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Telugu":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0C12:
                    if buffer.items[i+1].codepoint in [0x0C4C, 0x0C55]:
                        matched = True
            if buffer.items[i].codepoint in [0x0C3F, 0x0C46, 0x0C4A]:
                matched = 0x0C55 == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Kannada":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint in [0x0C89, 0x0C8B]:
                matched = 0x0CBE == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x0C92:
                matched = 0x0CCC == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Malayalam":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint in [0x0D07, 0x0D09]:
                matched = 0x0D57 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x0D0E:
                matched = 0x0D46 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x0D12:
                    if buffer.items[i+1].codepoint in [0x0D3E, 0x0D57]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Sinhala":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x0D85:
                    if buffer.items[i+1].codepoint in [0x0DCF, 0x0DD0, 0x0DD1]:
                        matched = True
            if buffer.items[i].codepoint in [0x0D8B, 0x0D8F, 0x0D94]:
                matched = 0x0DDF == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x0D8D:
                matched = 0x0DD8 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x0D91:
                    if buffer.items[i+1].codepoint in [0x0DCA, 0x0DD9, 0x0DDA, 0x0DDC, 0x0DDD, 0x0DDE]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Brahmi":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x11005:
                matched = 0x11038 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x1100B:
                matched = 0x1103E == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x1100F:
                matched = 0x11042 == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Khudawadi":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x112B0:
                    if buffer.items[i+1].codepoint in [0x112E0, 0x112E5, 0x112E6, 0x112E7, 0x112E8]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Tirhuta":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x11481:
                matched = 0x114B0 == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint in [0x1148B, 0x1148D]:
                matched = 0x114BA == buffer.items[i+1].codepoint
            if buffer.items[i].codepoint == 0x114AA:
                    if buffer.items[i+1].codepoint in [0x114B5, 0x114B6]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Modi":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint in [0x11600, 0x11601]:
                    if buffer.items[i+1].codepoint in [0x11639, 0x1163A]:
                        matched = True
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)
    if buffer.script == "Takri":
        i = 0
        while i < len(buffer.items)-1:
            matched = False
            if buffer.items[i].codepoint == 0x11680:
                    if buffer.items[i+1].codepoint in [0x116AD, 0x116B4, 0x116B5]:
                        matched = True
            if buffer.items[i].codepoint == 0x11686:
                matched = 0x116B2 == buffer.items[i+1].codepoint
            i = i + 1
            if matched: _insert_dotted_circle(buffer, i)



import fontTools.feaLib.ast as feaast
from fontFeatures.ttLib.Positioning import lookup_type


def glyphref(g):
    if len(g) == 1:
        return feaast.GlyphName(g[0])
    return feaast.GlyphClass([feaast.GlyphName(x) for x in sorted(g)])


def asFeaAST(self):
    lut = lookup_type(self)
    glyphs = [glyphref(x) for x in self.glyphs]
    positionings = list(zip(glyphs, self.valuerecords))
    if len(self.glyphs) == 1 or lut == 8:
        return feaast.SinglePosStatement(
            positionings,
            [glyphref(x) for x in self.precontext],
            [glyphref(x) for x in self.postcontext],
            (lut == 8),
        )
    if lut == 2:
        return feaast.PairPosStatement(
            glyphs[0], self.valuerecords[0], glyphs[1], self.valuerecords[1]
        )
    raise ValueError()



from .Routine import optimizations as routine_optimizations
from .FontFeatures import optimizations as overall_optimizations
import fontFeatures


class Optimizer:
    def __init__(self, ff):
        self.ff = ff

    def optimize(self, level=1):
        for r in self.ff.routines:
            self.optimize_routine(r, level)
            for k, v in self.ff.features.items():
                for n in v:
                    if isinstance(n, fontFeatures.Routine):
                        self.optimize_routine(n, level)

        for routinelist in self.ff.features.values():
            for r in routinelist:
                if isinstance(r, fontFeatures.Routine):
                    self.optimize_routine(r, level)

        for optimization in overall_optimizations:
            if level >= optimization.level:
                optimization().apply(self.ff)

    def optimize_routine(self, r, level):
        for optimization in routine_optimizations:
            if level >= optimization.level:
                optimization().apply(r, self.ff)



"""Routines for converting Positioning rules to and from XML."""

from lxml import etree
from fontFeatures import ValueRecord

# We can't monkeypatch this, because it's not our class.
def _valuerecord_fromXML(el):
    v = ValueRecord()
    for part in ["xPlacement", "yPlacement", "xAdvance", "yAdvance"]:
        if el.get(part):
            setattr(v, part, int(el.get(part)))
    return v


def _valuerecord_toXML(v):
    el = etree.Element("valuerecord")
    for part in ["xPlacement", "yPlacement", "xAdvance", "yAdvance"]:
        if getattr(v, part):
            el.set(part, str(getattr(v, part)))
    return el


def _toXML(self, root):
    self._makeglyphslots(root, "glyphs", self.glyphs)
    wrapper = etree.SubElement(root, "positions")
    for p in self.valuerecords:
        wrapper.append(_valuerecord_toXML(p))
    return root


@classmethod
def fromXML(klass, el):
    """Creates a rule from a lxml Element object."""
    position = el.find("positions")
    positions = [_valuerecord_fromXML(x) for x in position.findall("valuerecord")]
    rule = klass(
        klass._slotArray(klass, el.find("glyphs")),
        positions,
        precontext=klass._slotArray(klass, el.find("precontext")),
        postcontext=klass._slotArray(klass, el.find("postcontext")),
        address=el.get("address"),
        languages=el.get("languages"),
        flags=el.get("flags"),
    )
    if el.find("lookups"):
        rule.lookups = []
        for slot in list(el):
            routines = [Routine.fromXML(x) for x in slot.findall("routine")]
            rule.lookups.append(routines)

    return rule



from fontFeatures import Substitution, Routine, FontFeatures
from fontFeatures.optimizer import Optimizer

import unittest
import pytest


pytest.skip("Optimizer is currently broken", allow_module_level=True)

class TestSubstitution(unittest.TestCase):
    def test_MergeMultipleSingleSubstitutions_1(self):
        r1 = Routine(
            rules=[
                Substitution([["a", "y"]], [["b", "z"]]),
                Substitution([["b", "d", "a"]], [["c", "e", "f"]]),
            ]
        )
        Optimizer(FontFeatures()).optimize_routine(r1, level=1)
        self.assertEqual(r1.asFea(), "    sub [a y b d] by [c z c e];\n")

    def test_MergeMultipleSingleSubstitutions_2(self):
        r1 = Routine(
            rules=[
                Substitution([["a"]], [["b"]]),
                Substitution([["b"]], [["c"]]),
                Substitution([["d"]], [["e"]]),
                Substitution([["y"]], [["z"]]),
            ]
        )
        Optimizer(FontFeatures()).optimize_routine(r1, level=1)
        self.assertEqual(r1.asFea(), "    sub [a b d y] by [c c e z];\n")

    def test_GlyphClasses(self):
        r1 = Routine(
            rules=[Substitution([["a", "b", "c", "d", "e", "f", "g", "h"]], [["z"]]),]
        )
        ff = FontFeatures()
        Optimizer(ff).optimize_routine(r1, level=1)
        self.assertEqual(r1.asFea(), "    sub @class1 by z;\n")
        self.assertEqual(
            ff.namedClasses["class1"], ("a", "b", "c", "d", "e", "f", "g", "h")
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os
import glob

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

scripts = glob.glob("bin/*")

config = {
    'name': 'fontFeatures',
    'author': 'Simon Cozens',
    'author_email': 'simon@simon-cozens.org',
    'url': 'https://github.com/simoncozens/fontFeatures',
    'description': 'Python library for manipulating OpenType font features',
    'long_description': open('README.md', 'r').read(),
    'long_description_content_type': 'text/markdown',
    'license': 'MIT',
    'version': '1.7.4',
    'install_requires': install_requires,
    'extras_require': {
        'shaper': [
            "youseedee >=0.3.0",
            "babelfont >=3.0.0a1",
            "dataclasses ; python_version < '3.7'"
        ]
    },
    'classifiers': [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta"

    ],
    'package_dir': {'fontFeatures': 'fontFeatures'},
    'packages': find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"])
,
    'scripts': scripts,
    'zip_safe': False
}

if __name__ == '__main__':
    setup(**config)



import logging

__all__ = ["apply_to_buffer", "would_apply_at_position", "pre_post_context_matches", "_expand_slot"]


def i2s(buffer_items):
    return "|".join([x.glyph for x in buffer_items])

def _expand_slot(slot, namedclasses={}):
    expanded = []
    for g in slot:
        if g.startswith("@"):
            expanded.extend(namedclasses.get(g[1:], ""))
        else:
            expanded.append(g)
    return expanded

def glyphs_match(buffer_glyphs, routine_glyphs, namedclasses={}):
    if len(buffer_glyphs) != len(routine_glyphs):
        return False
    for a, b in zip(buffer_glyphs, routine_glyphs):
        if a.glyph not in _expand_slot(b, namedclasses):
            return False
    return True

def pre_post_context_matches(self, buf, ix, namedclasses={}):
    if hasattr(self, "precontext") and self.precontext:
        if ix < len(self.precontext):
            logging.getLogger("fontFeatures.shaperLib").debug(" - No, not enough precontext")
            return False
        precontext = buf[ix - len(self.precontext) : ix]
        if not glyphs_match(precontext, self.precontext, namedclasses):
            logging.getLogger("fontFeatures.shaperLib").debug(" - No, precontext doesn't match %s != %s" % (i2s(precontext), self.precontext))
            return False
    if hasattr(self, "postcontext") and self.postcontext:
        coverage = self.shaper_inputs()
        coverage_l = len(coverage)
        end_of_coverage = ix + coverage_l
        if end_of_coverage + len(self.postcontext) > len(buf):
            logging.getLogger("fontFeatures.shaperLib").debug(" - No, not enough postcontext")
            return False
        postcontext = buf[end_of_coverage : end_of_coverage + len(self.postcontext)]
        if not glyphs_match(postcontext, self.postcontext, namedclasses):
            logging.getLogger("fontFeatures.shaperLib").debug(" - No, postcontext doesn't match %s != %s" % (i2s(postcontext), self.postcontext))
            return False
    return True

def would_apply_at_position(self, buf, ix, namedclasses={}):
    logging.getLogger("fontFeatures.shaperLib").debug("Testing if rule would apply at position %i" % (ix))
    coverage = self.shaper_inputs()
    coverage_l = len(coverage)
    if coverage_l < 1: return False
    buffer_glyphs = buf[ix : ix + coverage_l]

    if not glyphs_match(buffer_glyphs, coverage, namedclasses):
        logging.getLogger("fontFeatures.shaperLib").debug(" - No! %s != %s" % (i2s(buffer_glyphs), coverage))
        return False

    if not pre_post_context_matches(self, buf, ix, namedclasses):
        return False

    logging.getLogger("fontFeatures.shaperLib").debug(" - Yes! %s == %s" % (i2s(buffer_glyphs), coverage))
    return True




#!/usr/bin/env python3

"""Generator of the function to prohibit certain vowel sequences.

It creates ``_hb_preprocess_text_vowel_constraints``, which inserts dotted
circles into sequences prohibited by the USE script development spec.
This function should be used as the ``preprocess_text`` of an
``hb_ot_complex_shaper_t``.

usage: ./gen-vowel-constraints.py ms-use/IndicShapingInvalidCluster.txt

"""

import collections
import youseedee

def write (s):
    sys.stdout.flush ()
    sys.stdout.buffer.write (s.encode ('utf-8'))
import sys

if len (sys.argv) != 2:
    sys.exit (__doc__)

script_order = {}
scripts = {}

for start, end,script in youseedee.parse_file_ranges("Scripts.txt"):
    for u in range (start, end + 1):
        scripts[u] = script
    if script not in script_order:
        script_order[script] = start

class ConstraintSet (object):
    """A set of prohibited code point sequences.

    Args:
        constraint (List[int]): A prohibited code point sequence.

    """
    def __init__ (self, constraint):
        # Either a list or a dictionary. As a list of code points, it
        # represents a prohibited code point sequence. As a dictionary,
        # it represents a set of prohibited sequences, where each item
        # represents the set of prohibited sequences starting with the
        # key (a code point) concatenated with any of the values
        # (ConstraintSets).
        self._c = constraint

    def add (self, constraint):
        """Add a constraint to this set."""
        if not constraint:
            return
        first = constraint[0]
        rest = constraint[1:]
        if isinstance (self._c, list):
            if constraint == self._c[:len (constraint)]:
                self._c = constraint
            elif self._c != constraint[:len (self._c)]:
                self._c = {self._c[0]: ConstraintSet (self._c[1:])}
        if isinstance (self._c, dict):
            if first in self._c:
                self._c[first].add (rest)
            else:
                self._c[first] = ConstraintSet (rest)

    @staticmethod
    def _indent (depth):
        return ('    ' * depth)

    @staticmethod
    def _cp_accessor(index):
        if index:
            return "buffer.items[i+{}].codepoint".format(index)
        return "buffer.items[i].codepoint"

    def __str__ (self, index=0, depth=2):
        s = []
        indent = self._indent (depth)

        if isinstance (self._c, list):
            if len (self._c) == 0:
                assert index == 2, 'Cannot use `matched` for this constraint; the general case has not been implemented'
                s.append ('{}matched = True\n'.format (indent))
            elif len (self._c) == 1:
                assert index == 1, 'Cannot use `matched` for this constraint; the general case has not been implemented'
                s.append ('{}matched = 0x{:04X} == {}\n'.format (indent, next (iter (self._c)), self._cp_accessor(index)))
            else:
                s.append ('{}if (0x{:04X} == {} and\n'.format (indent, self._c[0], self._cp_accessor(index)))
                if index:
                    s.append ('{}i + {} < len(buffer.items)-1 and\n'.format (self._indent (depth + 2), index + 1))
                for i, cp in enumerate (self._c[1:], start=1):
                    s.append ('{}0x{:04X} == {}{}\n'.format (
                        self._indent (depth + 2), cp, self._cp_accessor(index + i), '):' if i == len (self._c) - 1 else 'and')
                    )
                s.append ('{}matched = True\n'.format (self._indent (depth + 1)))
        else:
            cases = collections.defaultdict (set)
            for first, rest in sorted (self._c.items ()):
                cases[rest.__str__ (index + 1, depth + 2)].add (first)
            for body, labels in sorted (cases.items (), key=lambda b_ls: sorted (b_ls[1])[0]):
                if len(labels) == 1:
                    s.append (self._indent (depth + 1) + "if {} == 0x{:04X}:\n".format(self._cp_accessor(index), list(labels)[0]))
                else:
                    points = ", ".join(['0x{:04X}'.format(cp) for cp in sorted(labels)])
                    s.append (self._indent (depth + 1) + "if {} in [{}]:\n".format(self._cp_accessor(index), points))
                s.append (body)
        return ''.join (s)

constraints = {}
with open (sys.argv[1], encoding='utf-8') as f:
    constraints_header = []
    while True:
        line = f.readline ().strip ()
        if line == '#':
            break
        constraints_header.append(line)
    for line in f:
        j = line.find ('#')
        if j >= 0:
            line = line[:j]
        constraint = [int (cp, 16) for cp in line.split (';')[0].split ()]
        if not constraint: continue
        assert 2 <= len (constraint), 'Prohibited sequence is too short: {}'.format (constraint)
        script = scripts[constraint[0]]
        if script in constraints:
            constraints[script].add (constraint)
        else:
            constraints[script] = ConstraintSet (constraint)
        assert constraints, 'No constraints found'

print ('# The following functions are generated by running:')
print ('# %s ms-use/IndicShapingInvalidCluster.txt' % sys.argv[0])

print("""
from fontFeatures.shaperLib.Buffer import BufferItem

DOTTED_CIRCLE = 0x25CC

def _insert_dotted_circle(buf, index):
    dotted_circle = BufferItem.new_unicode(DOTTED_CIRCLE)
    buf.items.insert(index, dotted_circle)

""")

print ('def preprocess_text_vowel_constraints(buffer):')

for script, constraints in sorted (constraints.items (), key=lambda s_c: script_order[s_c[0]]):
    print(f'    if buffer.script == "{script}":')
    print ('        i = 0')
    print ('        while i < len(buffer.items)-1:')
    print ('            matched = False')
    write (str (constraints))
    print ('            i = i + 1')
    print ('            if matched: _insert_dotted_circle(buffer, i)')

